# Generated by Django 4.2.4 on 2023-09-07 03:29

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0014_schemalabel_schemadescription"),
    ]

    operations = [
        migrations.AlterField(
            model_name="formkitschema",
            name="label",
            field=models.CharField(
                blank=True,
                default=uuid.uuid4,
                help_text="Used as a human-readable label",
                max_length=1024,
                null=True,
                unique=True,
            ),
        ),
    ]



# ruff: noqa: F401 F811
# flake8: noqa: F401 F811

from http import HTTPStatus

import pytest
from django.test import Client
from django.urls import reverse
from pytest_django.fixtures import admin_client

from formkit_ninja.api import FormKitNodeIn


@pytest.mark.django_db
def test_node_create(admin_client: Client):
    """
    Test creation of Formkit components through the API
    """
    path = reverse("api-1.0.0:create_or_update_node")
    # Add a group node
    # This is a 'partisipa' type group node with
    # an icon and a label
    group = FormKitNodeIn(
        **{"$formkit": "group", "icon": "fa fa-user", "label": "Partisipa"}
    )
    data = group.model_dump_json(exclude_none=True)

    response = admin_client.post(
        path=path,
        data=data,
        content_type="application/json",
    )
    assert response.status_code == HTTPStatus.OK
    parent_reponse_json = response.json()
    field = FormKitNodeIn(
        parent_id=parent_reponse_json["key"],
        **{"$formkit": "text", "label": "Name of my Input"},
    )

    node_post = admin_client.post(
        path=path,
        data=field.model_dump_json(),
        content_type="application/json",
    )
    assert node_post.status_code == HTTPStatus.OK
    # Creating a new Node should return the parent and the node
    node_post_json = node_post.json()
    assert node_post_json["node"]["name"] == "name_of_my_input"

    # This provides us with a UUID for testing
    node_uuid = node_post_json["key"]

    # If we post this again, we should have another input with a deconflicted name
    field_2 = FormKitNodeIn(
        parent_id=parent_reponse_json["key"],
        **{"$formkit": "text", "label": "Name of my Input"},
    )

    node_post_2 = admin_client.post(
        path=path,
        data=field_2.model_dump_json(exclude_none=True),
        content_type="application/json",
    )
    assert node_post_2.json()["node"]["name"] == "name_of_my_input_1"

    # If we post this with its ID it should be recognized as an update
    # and the name should not change
    field_3 = FormKitNodeIn(
        parent_id=parent_reponse_json["key"],
        uuid=node_uuid,
        **{"$formkit": "text", "label": "Name of my Input"},
    )
    node_post_3 = admin_client.post(
        path=path,
        data=field_3.model_dump_json(exclude_none=True),
        content_type="application/json",
    )
    assert node_post_3.json()["node"]["name"] == "name_of_my_input"

    # If the name already exists and the label changes the name should not change
    field_4 = FormKitNodeIn(
        parent_id=parent_reponse_json["key"],
        uuid=node_uuid,
        label="New Label",
        **{"$formkit": "text"},
    )
    node_post_4 = admin_client.post(
        path=path,
        data=field_4.model_dump_json(exclude_none=True),
        content_type="application/json",
    )
    assert node_post_4.json()["node"]["name"] == "name_of_my_input"


@pytest.mark.django_db
def test_textarea_create(admin_client: Client):
    """
    Test creation of a Formkit TextArea node through the API
    """
    path = reverse("api-1.0.0:create_or_update_node")
    # Add a group node
    # This is a 'partisipa' type group node with
    # an icon and a label
    group = FormKitNodeIn(**{"$formkit": "textarea"})
    data = group.model_dump_json(exclude_none=True)

    response = admin_client.post(
        path=path,
        data=data,
        content_type="application/json",
    )
    assert response.status_code == HTTPStatus.OK



from typing import Literal

from pydantic import BaseModel, Field


class TestNodeType(BaseModel): ...


class TestTextNode(TestNodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["text"] = Field(default="text", alias="$formkit")


class TestGroupNode(BaseModel):
    children: list[TestTextNode]



from factory import Faker
from factory.django import DjangoModelFactory

from formkit_ninja import models


class FormKitSchemaNodeFactory(DjangoModelFactory):
    class Meta:
        model = models.FormKitSchemaNode

    # Assuming FormKitSchemaNode has these fields
    node = Faker("pydict")
    label = Faker("sentence")
    node_type = Faker("word")



# Generated by Django 4.2.4 on 2023-09-04 16:30

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0013_formkitschemanode_option_group"),
    ]

    operations = [
        migrations.CreateModel(
            name="SchemaLabel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.CharField(max_length=1024)),
                (
                    "lang",
                    models.CharField(
                        choices=[
                            ("en", "English"),
                            ("tet", "Tetum"),
                            ("pt", "Portugese"),
                        ],
                        default="en",
                        max_length=4,
                    ),
                ),
                (
                    "schema",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.formkitschema",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SchemaDescription",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.CharField(max_length=1024)),
                (
                    "lang",
                    models.CharField(
                        choices=[
                            ("en", "English"),
                            ("tet", "Tetum"),
                            ("pt", "Portugese"),
                        ],
                        default="en",
                        max_length=4,
                    ),
                ),
                (
                    "schema",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.formkitschema",
                    ),
                ),
            ],
        ),
    ]



import pgtrigger

NODE_CHANGE_ID = "formkitschemanode_change_id"
NODE_CHILDREN_CHANGE_ID = "nodechildren_change_id"


def create_sequence(sequence_name: str):
    return f"CREATE SEQUENCE IF NOT EXISTS {sequence_name};"


def drop_sequence(sequence_name: str):
    return f"DROP SEQUENCE IF EXISTS {sequence_name};"


def create_sequence_migration(sequence_name: str):
    return (create_sequence(sequence_name), drop_sequence(sequence_name))


def update_group_trigger(order_by_field: str, id_field: str = "id"):
    """
    Takes a model with an "order" field and
    a "group" field and adds a trigger to
    keep the ordering correct
    This assumes a pk field named "id" too
    """
    return pgtrigger.Trigger(
        name="order_on_update_option",
        when=pgtrigger.After,
        operation=pgtrigger.Update,
        func=pgtrigger.Func(
            f"""
                -- Do not allow a "null" value
                -- This stops Django from dumbly updating
                -- which can break the trigger
                if NEW."order" IS NULL then
                    NEW."order" = OLD."order";
                end if;
                if NEW."order" > OLD."order" then
                    update {{meta.db_table}}
                    set "order" = "order"- 1
                    where "order" <= NEW."order"
                    and "order" > OLD."order"
                    and "{order_by_field}" = NEW."{order_by_field}"
                    and "{id_field}" <> NEW."{id_field}";
                else
                    update {{meta.db_table}}
                    set "order" = "order"+ 1
                    where "order" >= NEW."order"
                    and "order" < OLD."order"
                    and "{order_by_field}" = NEW."{order_by_field}"
                    and "{id_field}" <> NEW."{id_field}";
                end if;
                RETURN NEW;
        """
        ),
        condition=pgtrigger.Condition(
            "pg_trigger_depth() = 0"
        ),  # Prevents infinite recursion
    )


def insert_group_trigger(order_by_field: str):
    return pgtrigger.Trigger(
        name="order_on_insert_option",
        when=pgtrigger.Before,
        operation=pgtrigger.Insert,
        func=pgtrigger.Func(
            f'NEW."order" = (SELECT coalesce(max("order"), 0) + 1 FROM {{meta.db_table}} WHERE {{meta.db_table}}."{order_by_field}" = NEW."{order_by_field}"); RETURN NEW;'
        ),
    )


def update_or_insert_group_trigger(order_by_field: str, id_field: str = "id"):
    return [
        update_group_trigger(order_by_field, id_field),
        insert_group_trigger(order_by_field),
    ]


def bump_sequence_value(
    value_field: str = "track_change", sequence_name: str = NODE_CHANGE_ID
):
    """
    Increment a sequence value and set the field referred to to that value.
    Intended to track latest changes across a model / multiple models
    to make syncing easier
    Remember to add the appropriate "CREATE SEQUENCE" code to a migration.
    That would be something like

    migrations.RunSQL(
        '''
        CREATE SEQUENCE IF NOT EXISTS ida_options_version_seq;
        ''',
        '''
        DROP SEQUENCE IF EXISTS ida_options_version_seq;
        ''',
    )
    """
    return pgtrigger.Trigger(
        name="version_on_update",
        when=pgtrigger.Before,
        operation=pgtrigger.Update | pgtrigger.Insert,
        func=f"""NEW."{value_field}" = nextval('{sequence_name}'); RETURN NEW;""",
    )



from importlib.util import find_spec

import pytest
from django.contrib.auth.models import User
from pytest_django.fixtures import live_server_helper

from formkit_ninja.parser.type_convert import NodePath
from formkit_ninja.schemas import Schemas
from tests.test_jinja_template import get_env

if find_spec("playwright") and find_spec("pytest_playwright"):
    from pytest_playwright.pytest_playwright import page  # type: ignore
else:
    # This is a workaround for the fact that pytest-playwright is not installed
    # This is a dummy fixture
    @pytest.fixture
    def page():
        yield None


"""
These are the forms used in Partisipa
To recreate, from the Partisipa project
(note we will deprecate 'ts' type form schemas soon)

from formkit_python_sync.get_schemas import _run_tsx
import tempfile
import json
for form_name, form_defn in _run_tsx().items():
     import tempfile
     with tempfile.NamedTemporaryFile(prefix=form_name, delete=False, mode='w') as o:
         o.write(json.dumps(form_defn))
         print(o.name)


"""

schemas = Schemas()


@pytest.fixture
def CFM_12_FF_12():
    return schemas.as_json("CFM_12_FF_12")


@pytest.fixture
def CFM_2_FF_4():
    return schemas.as_json("CFM_2_FF_4")


@pytest.fixture
def FF_14():
    return schemas.as_json("FF_14")


@pytest.fixture
def POM_1():
    return schemas.as_json("POM_1")


@pytest.fixture
def SF_1_1():
    return schemas.as_json("SF_1_1")


@pytest.fixture
def SF_1_2():
    return schemas.as_json("SF_1_2")


@pytest.fixture
def SF_1_3():
    return schemas.as_json("SF_1_3")


@pytest.fixture
def SF_2_3():
    return schemas.as_json("SF_2_3")


@pytest.fixture
def SF_4_1():
    return schemas.as_json("SF_4_1")


@pytest.fixture
def SF_4_2():
    return schemas.as_json("SF_4_2")


@pytest.fixture
def SF_6_2():
    return schemas.as_json("SF_6_2")


@pytest.fixture
def TF_13_2_1():
    return schemas.as_json("TF_13_2_1")


@pytest.fixture
def TF_6_1_1():
    return schemas.as_json("TF_6_1_1")


@pytest.fixture
def el_priority():
    """
    This represents a more complicated "el" element type
    """
    return {
        "$el": "div",
        "children": [
            "Priority ",
            {
                "$el": "span",
                "attrs": {"class": "ml-1"},
                "children": ["$: ($index + 1)"],
            },
        ],
        "attrs": {"class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"},
    }


@pytest.fixture
def simple_text_node():
    return {"$el": "span", "children": "Priority"}


@pytest.fixture
def formkit_text_node():
    return {
        "key": "activity_type",
        "id": "activity_type",
        "name": "activity_type",
        "label": "$pgettext('activity_type', 'Meeting or Training')",
        "$formkit": "select",
        "placeholder": '$gettext("Please select")',
        "class": "red",
        "options": [
            {"value": "1", "label": "Training"},
            {"value": "2", "label": "Meeting"},
        ],
    }


@pytest.fixture
def nested_formkit_text_node():
    return {
        "$formkit": "group",
        "children": [
            {
                "key": "activity_type",
                "id": "activity_type",
                "name": "activity_type",
                "label": "$pgettext('activity_type', 'Meeting or Training')",
                "$formkit": "select",
                "placeholder": '$gettext("Please select")',
                "class": "red",
                "options": [
                    {"value": "1", "label": "Training"},
                    {"value": "2", "label": "Meeting"},
                ],
            }
        ],
    }


@pytest.fixture()
def admin_page(page, live_server: live_server_helper.LiveServer, admin_user: User):
    if page is None:
        yield
    page.goto(f"{live_server.url}/admin", timeout=10000)
    page.get_by_label("Username:").fill(admin_user.username)
    page.get_by_label("Password:").fill("password")
    page.get_by_role("button", name="Log in").click()
    page.context.set_default_timeout(5000)
    yield page


@pytest.fixture()
def number_node():
    return NodePath.from_obj(
        {
            "$formkit": "number",
            "name": "foonum",
        }
    )


@pytest.fixture()
def group_node():
    return NodePath.from_obj(
        {
            "$formkit": "group",
            "name": "foo",
            "children": [
                {
                    "$formkit": "number",
                    "name": "foonum",
                }
            ],
        }
    )


@pytest.fixture()
def nested_group_node():
    return NodePath.from_obj(
        {
            "$formkit": "group",
            "name": "bar",
            "children": [
                {
                    "$formkit": "group",
                    "name": "foo",
                    "children": [
                        {
                            "$formkit": "number",
                            "name": "foonum",
                        }
                    ],
                }
            ],
        }
    )


@pytest.fixture()
def nested_repeater_node():
    return NodePath.from_obj(
        {
            "$formkit": "group",
            "name": "bar",
            "children": [
                {
                    "$formkit": "repeater",
                    "name": "foo",
                    "children": [
                        {
                            "$formkit": "number",
                            "name": "foonum",
                        }
                    ],
                }
            ],
        }
    )


@pytest.fixture()
def django_class_template():
    env = get_env()
    template = env.get_template("model.jinja2")
    return template


@pytest.fixture()
def admin_template():
    env = get_env()
    template = env.get_template("admin.jinja2")
    return template


@pytest.fixture()
def admin_py_template():
    """Test the entire admin file including import header"""
    env = get_env()
    template = env.get_template("admin.py.jinja2")
    return template


@pytest.fixture()
def api_template():
    env = get_env()
    template = env.get_template("api.jinja2")
    return template


@pytest.fixture()
def schema_out_template():
    env = get_env()
    template = env.get_template("schema.jinja2")
    return template


@pytest.fixture()
def pydantic_class_template():
    env = get_env()
    template = env.get_template("basemodel.jinja2")
    return template



from __future__ import annotations

import logging
from typing import (Annotated, Any, Dict, List, Literal, Self, TypedDict,
                    TypeVar, Union)
import warnings

from pydantic import (BaseModel, ConfigDict, Discriminator, Field,
                      PlainValidator, RootModel, Tag, field_serializer, model_serializer,
                      model_validator)
from pydantic_core import PydanticSerializationUnexpectedValue
from pydantic.functional_validators import field_validator
from json5 import loads as js_load

"""
This is a port of selected parts of the FormKit schema
to Pydantic models.
"""


def parse_vue_syntax(key: str, value: str) -> tuple[str, Any] | tuple[None, None]:
    """
    Try to convert a Vue key,value to valid JSON
    """

    def try_convert(v):
        try:
            return js_load(v)
        except ValueError:
            pass
        return v

    if key.startswith("v-bind"):
        return parse_vue_syntax(key.replace("v-bind", ":"), value)
    if key.startswith(":"):
        return key[1:], try_convert(value)
    return None, None

def get_additional_props(
    object_in: dict[str, Any], exclude: set[str] = set()
) -> dict[str, Any] | None:
    """
    Parse the object or database return (dict)
    to break out fields we handle in JSON

    A FormKit node can have 'arbitrary' additional properties
    For instance classes to apply to child nodes
    here we can't realistically cover every scenario so
    fall back to JSON storage for thes

    However: if we're coming from the database we already store these in a separate field
    """
    props: dict[str, Any] = {**object_in.get("additional_props", {})}
    props.update({k: object_in[k] for k in object_in.keys() - exclude})
    return props if props else None


logger = logging.getLogger(__name__)

HtmlAttrs = Dict[str, str | Dict[str, str]]


# Radio, Select, Autocomplete and Dropdown nodes have
# these options
OptionsType = str | list[dict[str, Any]] | list[str] | Dict[str, str] | None


class FormKitSchemaCondition(BaseModel):
    node_type: Literal["condition"] = Field(default="condition", exclude=True)
    # If, Then, Else are aliases bacause they're reserved words
    if_condition: str = Field(..., alias="if")
    then_condition: Node | List[Node] = Field(..., alias="then")
    else_condition: Node | List[Node] | None = Field(None, alias="else")


class FormKitSchemaMeta(RootModel):
    root: dict[str, str | float | int | bool | None]


class FormKitTypeDefinition(BaseModel): ...


class FormKitContextShape(BaseModel):
    type: Literal["input", "list", "group"]
    value: Any = None
    _value: Any = None


class FormKitListValue(RootModel):
    root: str | list[str] | list[dict[str, str]]


class FormKitListStatement(RootModel):
    """
    A full loop statement in tuple syntax. Can be read like "foreach value, key? in list"
    A 2 or 2 element tuple of value, key, and list or value, list
    """

    root: tuple[str, float | int | str, list[FormKitListValue]]


class FormKitSchemaAttributesCondition(BaseModel):
    if_: str = Field(alias="if")
    then_: FormKitAttributeValue = Field(alias="then")
    else_: FormKitAttributeValue | None = Field(None, alias="else")
    model_config = ConfigDict(populate_by_name=True)


class FormKitAttributeValue(RootModel):
    """
    The possible value types of attributes (in the schema)
    """

    root: (
        str
        | int
        | float
        | bool
        | None
        | FormKitSchemaAttributes
        | FormKitSchemaAttributesCondition
    )


class FormKitSchemaAttributes(RootModel):
    root: dict[
        str,
        FormKitAttributeValue
        | FormKitSchemaAttributes
        | FormKitSchemaAttributesCondition,
    ]


def child_validate(value: Any):
    if isinstance(value, list):
        models = [DiscriminatedNodeType.model_validate(x).root for x in value]
        return models
    return value


# def child_validate(
#     value: Any, handler: ValidatorFunctionWrapHandler, _info: ValidationInfo
# ) -> Any:
#     """Simplify the error message to avoid a gross error stemming
#     from exhaustive checking of all union options.
#     """
#     try:
#         value = handler(value)
#         print(value)
#         return value

#     except ValidationError:
#         raise PydanticCustomError(
#             'invalid_json',
#             'Input is not valid json',
#         )


class FormKitSchemaProps(BaseModel):
    """
    Properties available in all schema nodes.
    """

    children: Annotated[
        str
        | list[
            FormKitType
            | FormKitSchemaDOMNode
            | FormKitSchemaComponent
            | FormKitSchemaCondition
        ]
        | None,
        PlainValidator(child_validate),
    ] = Field(None)
    key: str | None = None
    if_condition: str | None = Field(None, alias="if")
    for_loop: FormKitListStatement | None = Field(None, alias="for")
    bind: str | None = None
    meta: FormKitSchemaMeta | None = None

    # These are not formal parts of spec, but
    # are attributes defined in ts as Record<string, any>
    # id: str | uuid.UUID | None = Field(None)
    id: str | None = Field(None)
    name: str | None = Field(None)
    label: str | None = Field(None)
    help: str | None = Field(None)
    validation: str | None = Field(None)
    validationLabel: str | None = Field(None, alias="validation-label")
    validationVisibility: str | None = Field(None, alias="validation-visibility")
    validationMessages: Annotated[
        Dict[str, str] | str | None, Field(None, alias="validation-messages")
    ]
    placeholder: str | None = Field(None)
    value: str | None = Field(None)
    prefixIcon: str | None = Field(None, alias="prefix-icon")
    classes: Annotated[str | Dict[str, str] | None, Field(None)]

    # FormKit allows arbitrary values, we do our best to represent these here
    # Additional Props can be quite a complicated structure
    additional_props: None | Dict[str, str | int | Dict[str, Any]] = Field(
        None, validate_default=True
    )
    model_config = ConfigDict(populate_by_name=True)

    @classmethod
    def exclude_fields(cls) -> set[str]:
        """
        We want to exclude from "additional_props" any real fields on the model.
        We do this by "annotation" as we're expecting to have this by alias
        """
        exclude: set[str] = set()
        for field_name, field_obj in cls.model_fields.items():
            exclude.add(field_name)
            if field_obj.alias:
                exclude.add(field_obj.alias)
        return exclude

    # @model_validator(mode='wrap')
    # @classmethod
    # def log_failed_validation(cls, data: Any, handler: ModelWrapValidatorHandler[Self]) -> Self:
    #     try:
    #         validated = handler(data)
    #     except ValidationError:
    #         logging.error('Model %s failed to validate with data %s', cls, data)
    #         raise
    #     return validated

    @model_validator(mode="before")
    @classmethod
    def pre_validate_run(cls, data: Any) -> Any:
        """
        Before we "validate", we're going to
        apply some transforms to the data
        """

        if not isinstance(data, dict):
            # This should only apply on inputs of this type
            return data

        node_type = get_node_type(data)

        if node_type == "formkit":
            additional_props = get_additional_props(data, exclude=cls.exclude_fields())
            return {
                **data,
                "additional_props": additional_props,
                "node_type": node_type,
            }

        return {**data, "node_type": node_type}

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # We get a storm of "UnexpectedValueWarnings"
            result = handler(self)
        # return result
        if "additional_props" in result:
            if result["additional_props"] is not None:
                result.update(result["additional_props"])
            del result["additional_props"]

        # HTML attribute handling
        # When an input HTML has an attribute beginning with a colon (:)
        # This is a `v-bind` attribute
        # This can be a variable, Javascript expression, or function.
        # We can't support all of this (nor do we want to probably)
        # but we can support the basic variable binding
        renamed: dict[str, Any] = {}
        renamed_fields: set[str] = set()
        if isinstance(result, dict):
            k: str
            for k, v in result.items():
                _k, _v = parse_vue_syntax(k, v)
                if _k:
                    renamed[_k] = _v
                    renamed_fields.add(k)
            result.update(renamed)
        for k in renamed_fields:
            result.pop(k)

        # "@submit" or "@click" cannot be serialized well, they're probably best
        # handled client side, at least for now
        for unparseable in [k for k in result.keys() if k.startswith("@")]:
            result.pop(unparseable)

        return result

    @field_validator("validationMessages")
    @classmethod
    def validation_messages(cls, value, _info):
        """
        When we import from HTML we may have a string
        However this may be JSON so coerce to a JSON object
        if it smells like JSON
        """
        return value    


# We defined this after the model above as it's a circular reference
ChildNodeType = str | list[FormKitSchemaProps | str] | FormKitSchemaCondition | None


class NodeType(FormKitSchemaProps):
    """
    Abstract for "node" types
    """

    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    readonly: bool | None = None

    @model_validator(mode="before")
    @classmethod
    def pre_validate_run(cls, data: Any) -> Any:
        """
        Before we "validate", we're going to
        apply some transforms to the data
        """

        if not isinstance(data, dict):
            # This should only apply on inputs of this type
            return data

        if issubclass(cls, NodeType) and ("$formkit" in data or data == {}):
            node_type = "formkit"
        elif "$el" in data:
            node_type = "element"
        elif "$cmp" in data:
            node_type = "component"
        elif "$formkit" in data:
            node_type = "formkit"
        elif "if" in data:
            node_type = "condition"
        else:
            # Fallback: assume it is a NodeType
            # We need a better way to handle this
            node_type = "formkit"

        additional_props = get_additional_props(data, exclude=cls.exclude_fields())
        return {
            **data,
            "additional_props": additional_props,
            "node_type": node_type,
            "formkit": cls.model_fields["formkit"].default,
        }


class FormNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["form"] = Field(default="form", alias="$formkit")


class TextNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["text"] = Field(default="text", alias="$formkit")
    text: str | None = None


class TextAreaNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Annotated[Literal["textarea"], Field(alias="$formkit")]
    text: str | None = None


class DateNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["date"] = Field(default="date", alias="$formkit")


class CurrencyNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["currency"] = Field(default="currency", alias="$formkit")


class UuidNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["uuid"] = Field(default="uuid", alias="$formkit")


class DatePickerNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["datepicker"] = Field(default="datepicker", alias="$formkit")
    calendarIcon: str = "calendar"
    format: str = "DD/MM/YY"
    nextIcon: str = "angleRight"
    prevIcon: str = "angleLeft"


class CheckBoxNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["checkbox"] = Field(default="checkbox", alias="$formkit")


class NumberNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["number"] = Field(default="number", alias="$formkit")
    text: str | None = None
    max: int | None = None
    min: int | None = None
    step: str | None = None


class PasswordNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["password"] = Field(default="password", alias="$formkit")
    name: str | None = None


class HiddenNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["hidden"] = Field(default="hidden", alias="$formkit")


class RadioNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["radio"] = Field(default="radio", alias="$formkit")
    name: str | None = None
    options: OptionsType = Field(None)


class SelectNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["select"] = Field(default="select", alias="$formkit")
    options: OptionsType = Field(None)


class AutocompleteNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["autocomplete"] = Field(default="autocomplete", alias="$formkit")
    options: OptionsType = Field(None)


class EmailNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["email"] = Field(default="email", alias="$formkit")


class TelNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["tel"] = Field(default="tel", alias="$formkit")


class DropDownNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["dropdown"] = Field(default="dropdown", alias="$formkit")
    options: OptionsType = Field(None)
    empty_message: str | None = Field(None, alias="empty-message")
    select_icon: str | None = Field(None, alias="selectIcon")
    placeholder: str | None = None


class RepeaterNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["repeater"] = Field(default="repeater", alias="$formkit")
    up_control: bool | None = Field(default=None, alias="upControl")
    down_control: bool | None = Field(default=None, alias="downControl")
    add_label: str | None = Field(default=None, alias="addLabel")
    name: str | None = None
    min: int | None = None
    max: int | None = None


class GroupNode(NodeType):
    node_type: Literal["formkit"] = Field(default="formkit", exclude=True)
    formkit: Literal["group"] = Field(default="group", alias="$formkit")
    text: str | None = None


# This is useful for "isinstance" checks
# which do not work with "Annotated" below
FormKitType = (
    FormNode
    | TextNode
    | TextAreaNode
    | CheckBoxNode
    | PasswordNode
    | SelectNode
    | AutocompleteNode
    | EmailNode
    | NumberNode
    | RadioNode
    | GroupNode
    | DateNode
    | DatePickerNode
    | DropDownNode
    | RepeaterNode
    | TelNode
    | CurrencyNode
    | HiddenNode
    | UuidNode
)

FormKitSchemaFormKit = Annotated[
    Union[
        FormNode,
        TextNode,
        TextAreaNode,
        CheckBoxNode,
        PasswordNode,
        SelectNode,
        AutocompleteNode,
        EmailNode,
        NumberNode,
        RadioNode,
        GroupNode,
        DateNode,
        DatePickerNode,
        DropDownNode,
        TelNode,
        CurrencyNode,
        HiddenNode,
        UuidNode,
        RepeaterNode,
    ],
    Field(discriminator="formkit"),
]


class FormKitSchemaDOMNode(FormKitSchemaProps):
    """
    HTML elements are defined using the $el property.
    You can use $el to render any HTML element.
    Attributes can be added with the attrs property,
    and content is assigned with the children property
    """

    node_type: Literal["element"] = Field(default="element", exclude=True)
    el: str = Field(alias="$el")
    attrs: FormKitSchemaAttributes | None = None
    model_config = ConfigDict(populate_by_name=True)


class FormKitSchemaComponent(FormKitSchemaProps):
    """
    Components can be defined with the $cmp property
    The $cmp property should be a string that references
    a globally defined component or a component passed
    into FormKitSchema with the library prop.
    """

    node_type: Literal["component"] = Field(default="component", exclude=True)

    cmp: str = Field(
        ...,
        alias="$cmp",
        description="The $cmp property should be a string that references a globally defined component or a component passed into FormKitSchema with the library prop.",  # noqa: E501
    )
    props: dict[str, str | Any] | None = None
    model_config = ConfigDict(populate_by_name=True)


Node = Annotated[
    Union[
        FormKitSchemaFormKit,
        FormKitSchemaDOMNode,
        FormKitSchemaComponent,
        FormKitSchemaCondition,
    ],
    Field(discriminator="node_type"),
]

# This necessary to properly "populate" some more complicated models
FormKitSchemaAttributesCondition.model_rebuild()
FormKitAttributeValue.model_rebuild()
FormKitSchemaDOMNode.model_rebuild()
FormKitSchemaCondition.model_rebuild()
FormKitSchemaComponent.model_rebuild()

FormKitSchemaDOMNode.model_rebuild()

Model = TypeVar("Model", bound="BaseModel")
StrBytes = str | bytes


NODE_TYPE = Literal["condition", "formkit", "element", "component"]
FORMKIT_TYPE = Literal[
    "form",
    "text",
    "textarea",
    "tel",
    "currency",
    "select",
    "checkbox",
    "number",
    "group",
    "list",
    "password",
    "button",
    "radio",
    "form",
    "date",
    "datepicker",
    "dropdown",
    "repeater",
    "autocomplete",
    "email",
    "uuid",
]


class Discriminators(TypedDict, total=False):
    node_type: NODE_TYPE
    formkit: FORMKIT_TYPE


def get_node_type(obj: dict) -> Literal["text", "element", "formkit", "component"]:
    """
    Pydantic requires nodes to be "differentiated" by a field value
    when used in a Union type situation.
    This function should return the 'node_type' values and if present 'Formkit' value
    which corresponds to the object being inspected.
    """
    if "root" in obj:
        return get_node_type(obj["root"])

    if isinstance(obj, str):
        return "text"

    if isinstance(obj, dict) and len(obj.keys()) == 0:
        return "text"

    if "$el" in obj:
        return "element"
    if "$formkit" in obj:
        return "formkit"
    if "$cmp" in obj:
        return "component"

    raise KeyError(f"Could not determine node type for {obj}")


NodeTypes = (
    FormKitType | FormKitSchemaDOMNode | FormKitSchemaComponent | FormKitSchemaCondition
)


class FormKitNode(RootModel):
    root: NodeTypes

    def model_dump(self, *args, **kwargs):
        # Set some sensible defaults for "to_dict"
        if "by_alias" not in kwargs:
            kwargs["by_alias"] = True
        if "exclude_none" not in kwargs:
            kwargs["exclude_none"] = True
        # self.model_dump(*args, **kwargs) fails on some instances (maybe
        # a pydantic bug?)
        _ = self.root.model_dump(*args, **kwargs)
        if "additional_props" in _:
            _.update(_["additional_props"])
            del _["additional_props"]
        return _


class FormKitSchema(RootModel):
    root: list[NodeTypes]

    @classmethod
    def model_validate(cls, obj: Any, *args, **kwargs) -> Self:
        if isinstance(obj, dict):
            obj = [obj]
        return super().model_validate(obj, *args, **kwargs)


FormKitSchemaDefinition = Node | list[Node] | FormKitSchemaCondition


def get_discriminator_v(v: Any) -> str:

    if isinstance(v, str):
        return "string"
    if isinstance(v, dict):
        if "$formkit" in v:
            return f"{v.get('$formkit')}"
        elif "$el" in v:
            return "element"
        elif "if" in v:
            return "condition"
        elif "$cmp" in v:
            return "component"

    if hasattr(v, "formkit"):
        return f"{v.formkit}"
    elif hasattr(v, "el"):
        return "element"
    elif hasattr(v, "if_condition"):
        return "condition"
    elif hasattr(v, "cmp"):
        return "element"

    raise AssertionError("Could not determine node type")


class DiscriminatedNodeType(RootModel):
    """
    This is intended to replace "FormKitNode"
    with a more efficient tagged discriminator
    """

    root: Annotated[
        (
            Annotated[FormNode, Tag("form")]
            | Annotated[TextNode, Tag("text")]
            | Annotated[TextAreaNode, Tag("textarea")]
            | Annotated[CheckBoxNode, Tag("checkbox")]
            | Annotated[PasswordNode, Tag("password")]
            | Annotated[SelectNode, Tag("select")]
            | Annotated[AutocompleteNode, Tag("autocomplete")]
            | Annotated[EmailNode, Tag("email")]
            | Annotated[NumberNode, Tag("number")]
            | Annotated[RadioNode, Tag("radio")]
            | Annotated[GroupNode, Tag("group")]
            | Annotated[DateNode, Tag("date")]
            | Annotated[DatePickerNode, Tag("datepicker")]
            | Annotated[DropDownNode, Tag("dropdown")]
            | Annotated[RepeaterNode, Tag("repeater")]
            | Annotated[TelNode, Tag("tel")]
            | Annotated[CurrencyNode, Tag("currency")]
            | Annotated[HiddenNode, Tag("hidden")]
            | Annotated[UuidNode, Tag("uuid")]
            # Additional types
            | Annotated[FormKitSchemaAttributesCondition, Tag("condition")]
            | Annotated[FormKitSchemaDOMNode, Tag("element")]
            | Annotated[FormKitSchemaCondition, Tag("condition")]
            | Annotated[FormKitSchemaComponent, Tag("component")]
            | Annotated[str, Tag("string")]
        ),
        Discriminator(get_discriminator_v),
    ]


class DiscriminatedNodeTypeSchema(RootModel):
    root: list[DiscriminatedNodeType]


FormKitSchema.model_rebuild()
FormKitSchemaCondition.model_rebuild()
PasswordNode.model_rebuild()
RepeaterNode.model_rebuild()



from formkit_ninja.error_in_inheritance import TestGroupNode, TestTextNode


def test_child_serializer():
    assert TestTextNode().model_dump(by_alias=True) == {"$formkit": "text"}
    group = TestGroupNode(children=[TestTextNode()])
    assert group.model_dump(by_alias=True) == {"children": [{"$formkit": "text"}]}



import pytest

from formkit_ninja.models import check_valid_django_id


def test_valid_id():
    assert check_valid_django_id("valid_id") is None


def test_invalid_id_starting_with_digit():
    with pytest.raises(TypeError):
        check_valid_django_id("1invalid")


def test_invalid_id_ending_with_underscore():
    with pytest.raises(TypeError):
        check_valid_django_id("invalid_")


def test_invalid_id_with_keyword():
    with pytest.raises(TypeError):
        check_valid_django_id("for")


def test_invalid_id_with_softkeyword():
    with pytest.raises(TypeError):
        check_valid_django_id("async")



# Generated by Django 4.2.1 on 2023-08-28 05:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("formkit_ninja", "0004_remove_formkitschema_key_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="option",
            name="last_updated",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AddField(
            model_name="option",
            name="object_id",
            field=models.IntegerField(
                blank=True,
                help_text="This is a reference to the primary key of the original source object (typically a PNDS ztable ID) or a user-specified ID for a new group",
                null=True,
            ),
        ),
        migrations.CreateModel(
            name="OptionLabel",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("label", models.CharField(max_length=1024)),
                (
                    "lang",
                    models.CharField(
                        choices=[
                            ("en", "English"),
                            ("tet", "Tetum"),
                            ("pt", "Portugese"),
                        ],
                        default="en",
                        max_length=4,
                    ),
                ),
                (
                    "option",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.option",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="OptionGroup",
            fields=[
                (
                    "group",
                    models.CharField(
                        help_text="The label to use for these options",
                        max_length=1024,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        blank=True,
                        help_text="This is a reference to the original source object (typically a PNDS ztable)",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
        migrations.AddConstraint(
            model_name="optionlabel",
            constraint=models.UniqueConstraint(
                fields=("option", "lang"), name="unique_option_label"
            ),
        ),
    ]



"""
Django settings for testproject project.

Generated by 'django-admin startproject' using Django 4.1.3.

For more information on this file, see
https://docs.djangoproject.com/en/4.1/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.1/ref/settings/
"""

import importlib
from importlib.util import find_spec
from os import environ
from pathlib import Path

from django.conf.locale import LANG_INFO
from django.utils.translation import gettext_lazy as _

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-uq^#m^-8-k+f%jag(2kv$p#45%o!#au!877xby@c_tu*ccdr%f"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "formkit_ninja",
    "ninja",
    "pgtrigger",
    "pghistory",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "testproject.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "testproject.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.1/ref/settings/#databases
# Note:
# If you have no database running yet
# maybe start one with `docker run -p 5432:5432 -e POSTGRES_HOST_AUTH_METHOD=trust postgres`
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": "postgres",
        "USER": "postgres",
        "HOST": "localhost",
    }
}


# Password validation
# https://docs.djangoproject.com/en/4.1/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.1/topics/i18n/

# Internationalization
# https://docs.djangoproject.com/en/4.1/topics/i18n/
EXTRA_LANG_INFO = {
    "tet": {
        "bidi": False,  # right-to-left
        "code": "tet",
        "name": "Tetum",
        "name_local": "Tetum",
    }
}
LANG_INFO.update(EXTRA_LANG_INFO)
LANGUAGE_CODE = environ.get("DJANGO_LANGUAGE_CODE", "en")

TIME_ZONE = "Asia/Dili"

LANGUAGES = [
    ("tet", _("Tetum")),
    ("en", _("English")),
    ("pt", _("Portuguese")),
]


TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.1/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.1/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

if DEBUG and find_spec("django_extensions"):
    # "pip install django-extensions"
    INSTALLED_APPS.append("django_extensions")

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": (
                "rich.logging.RichHandler"
                if importlib.util.find_spec("rich")
                else "logging.StreamHandler"
            ),
        }
    },
    "loggers": {
        "formkit_ninja": {"handlers": ["console"], "level": "DEBUG", "propagate": True},
    },
}



# Generated by Django 4.2.7 on 2023-11-11 04:25

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0023_formkitschemanodeevent_and_more"),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name="formkitschemanode",
            name="insert_insert",
        ),
        pgtrigger.migrations.RemoveTrigger(
            model_name="formkitschemanode",
            name="update_update",
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="protected",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="formkitschemanodeevent",
            name="protected",
            field=models.BooleanField(default=False),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="protect_deletes",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (OLD."protected")',
                    func="RAISE EXCEPTION 'pgtrigger: Cannot delete rows from % table', TG_TABLE_NAME;",
                    hash="4d74bdff98561d69e24a51782fd756b9410242b4",
                    operation="DELETE",
                    pgid="pgtrigger_protect_deletes_ec7d5",
                    table="formkit_ninja_formkitschemanode",
                    when="BEFORE",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="insert_insert",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='INSERT INTO "formkit_ninja_formkitschemanodeevent" ("additional_props", "created", "created_by_id", "description", "id", "is_active", "label", "node", "node_type", "option_group_id", "pgh_context_id", "pgh_created_at", "pgh_label", "pgh_obj_id", "protected", "text_content", "track_change", "updated", "updated_by_id") VALUES (NEW."additional_props", NEW."created", NEW."created_by_id", NEW."description", NEW."id", NEW."is_active", NEW."label", NEW."node", NEW."node_type", NEW."option_group_id", _pgh_attach_context(), NOW(), \'insert\', NEW."id", NEW."protected", NEW."text_content", NEW."track_change", NEW."updated", NEW."updated_by_id"); RETURN NULL;',
                    hash="c5bce62ecf2f791f422ee61d09f7cedeccee6637",
                    operation="INSERT",
                    pgid="pgtrigger_insert_insert_72c4c",
                    table="formkit_ninja_formkitschemanode",
                    when="AFTER",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="update_update",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (OLD.* IS DISTINCT FROM NEW.*)",
                    func='INSERT INTO "formkit_ninja_formkitschemanodeevent" ("additional_props", "created", "created_by_id", "description", "id", "is_active", "label", "node", "node_type", "option_group_id", "pgh_context_id", "pgh_created_at", "pgh_label", "pgh_obj_id", "protected", "text_content", "track_change", "updated", "updated_by_id") VALUES (NEW."additional_props", NEW."created", NEW."created_by_id", NEW."description", NEW."id", NEW."is_active", NEW."label", NEW."node", NEW."node_type", NEW."option_group_id", _pgh_attach_context(), NOW(), \'update\', NEW."id", NEW."protected", NEW."text_content", NEW."track_change", NEW."updated", NEW."updated_by_id"); RETURN NULL;',
                    hash="03fbb3ba0fbdd08226bdc6ab9bfc423dc993cf2b",
                    operation="UPDATE",
                    pgid="pgtrigger_update_update_d7c99",
                    table="formkit_ninja_formkitschemanode",
                    when="AFTER",
                ),
            ),
        ),
    ]



# Generated by Django 4.2.5 on 2023-09-10 16:44

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0015_alter_formkitschema_label"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="nodechildren",
            options={"ordering": ("parent_id", "order")},
        ),
        migrations.AlterField(
            model_name="formcomponents",
            name="order",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="nodechildren",
            name="order",
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="option",
            name="order",
            field=models.IntegerField(blank=True, null=True),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formcomponents",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_insert_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='NEW."order" = (SELECT coalesce(max("order"), 0) + 1 FROM formkit_ninja_formcomponents WHERE formkit_ninja_formcomponents."schema_id" = NEW."schema_id"); RETURN NEW;',
                    hash="32de6ecbd50ae312fcc482cdba82ce45e9b76908",
                    operation="INSERT",
                    pgid="pgtrigger_order_on_insert_option_1fbd2",
                    table="formkit_ninja_formcomponents",
                    when="BEFORE",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formcomponents",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_update_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (pg_trigger_depth() = 0)",
                    func='\n                -- Do not allow a "null" value\n                -- This stops Django from dumbly updating\n                -- which can break the trigger\n                if NEW."order" IS NULL then\n                    NEW."order" = OLD."order";\n                end if;\n                if NEW."order" > OLD."order" then\n                    update formkit_ninja_formcomponents\n                    set "order" = "order"- 1\n                    where "order" <= NEW."order"\n                    and "order" > OLD."order"\n                    and "schema_id" = OLD."schema_id"\n                    and "id" <> NEW."id";\n                else\n                    update formkit_ninja_formcomponents\n                    set "order" = "order"+ 1\n                    where "order" >= NEW."order"\n                    and "order" < OLD."order"\n                    and "schema_id" = OLD."schema_id"\n                    and "id" <> NEW."id";\n                end if;\n                RETURN NEW;\n        ',
                    hash="e87c38bfcb4fa1b13b8f6f2aa5cb186fe831c8a3",
                    operation="UPDATE",
                    pgid="pgtrigger_order_on_update_option_a5326",
                    table="formkit_ninja_formcomponents",
                    when="AFTER",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="nodechildren",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_insert_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='NEW."order" = (SELECT coalesce(max("order"), 0) + 1 FROM formkit_ninja_nodechildren WHERE formkit_ninja_nodechildren."parent_id" = NEW."parent_id"); RETURN NEW;',
                    hash="c66dfe1401c425e23a30d21ed7cd257cfc02b5b5",
                    operation="INSERT",
                    pgid="pgtrigger_order_on_insert_option_1816c",
                    table="formkit_ninja_nodechildren",
                    when="BEFORE",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="nodechildren",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_update_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (pg_trigger_depth() = 0)",
                    func='\n                -- Do not allow a "null" value\n                -- This stops Django from dumbly updating\n                -- which can break the trigger\n                if NEW."order" IS NULL then\n                    NEW."order" = OLD."order";\n                end if;\n                if NEW."order" > OLD."order" then\n                    update formkit_ninja_nodechildren\n                    set "order" = "order"- 1\n                    where "order" <= NEW."order"\n                    and "order" > OLD."order"\n                    and "parent_id" = OLD."parent_id"\n                    and "id" <> NEW."id";\n                else\n                    update formkit_ninja_nodechildren\n                    set "order" = "order"+ 1\n                    where "order" >= NEW."order"\n                    and "order" < OLD."order"\n                    and "parent_id" = OLD."parent_id"\n                    and "id" <> NEW."id";\n                end if;\n                RETURN NEW;\n        ',
                    hash="a421dc6da53fa50d3226e4bb4802866879344941",
                    operation="UPDATE",
                    pgid="pgtrigger_order_on_update_option_44228",
                    table="formkit_ninja_nodechildren",
                    when="AFTER",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="option",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_insert_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='NEW."order" = (SELECT coalesce(max("order"), 0) + 1 FROM formkit_ninja_option WHERE formkit_ninja_option."group_id" = NEW."group_id"); RETURN NEW;',
                    hash="eff471a7791c174bd5c48d6ebb6baad36aa108d9",
                    operation="INSERT",
                    pgid="pgtrigger_order_on_insert_option_e1443",
                    table="formkit_ninja_option",
                    when="BEFORE",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="option",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_update_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (pg_trigger_depth() = 0)",
                    func='\n                -- Do not allow a "null" value\n                -- This stops Django from dumbly updating\n                -- which can break the trigger\n                if NEW."order" IS NULL then\n                    NEW."order" = OLD."order";\n                end if;\n                if NEW."order" > OLD."order" then\n                    update formkit_ninja_option\n                    set "order" = "order"- 1\n                    where "order" <= NEW."order"\n                    and "order" > OLD."order"\n                    and "group_id" = OLD."group_id"\n                    and "id" <> NEW."id";\n                else\n                    update formkit_ninja_option\n                    set "order" = "order"+ 1\n                    where "order" >= NEW."order"\n                    and "order" < OLD."order"\n                    and "group_id" = OLD."group_id"\n                    and "id" <> NEW."id";\n                end if;\n                RETURN NEW;\n        ',
                    hash="ea7dde4b2c32fe380c79eaf105621171c396fa71",
                    operation="UPDATE",
                    pgid="pgtrigger_order_on_update_option_417cb",
                    table="formkit_ninja_option",
                    when="AFTER",
                ),
            ),
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-16 14:08

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0020_remove_formcomponents_order_on_update_option_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="formkitschemanode",
            name="is_active",
            field=models.BooleanField(default=True),
        ),
    ]



# ruff: noqa: F401 F811
# flake8: noqa: F401 F811

import os
from importlib.util import find_spec

import pytest
from playwright.sync_api import Page

from formkit_ninja import models

os.environ.setdefault("DJANGO_ALLOW_ASYNC_UNSAFE", "true")

playwright = pytest.mark.skipif(
    # 'playwright' is not installed
    find_spec("playwright") is None or find_spec("pytest_playwright") is None,
    reason="playwright is not installed",
)


@playwright
def test_home_page(admin_page):
    admin_page.get_by_role("link", name="Form kit schema nodes").click()
    admin_page.get_by_role("link", name="Add form kit schema node").click()
    # There was a bug identified by this test where "label" in JSON
    # would override "label" in the parent model
    admin_page.get_by_label("Label:").fill("test")
    admin_page.get_by_label("Node type:").select_option("$formkit")
    admin_page.get_by_label("Description:").fill("This is a test node")
    admin_page.get_by_role("button", name="Save and continue editing").click()
    admin_page.get_by_label("Name:").fill("test")
    admin_page.get_by_label("Key:").fill("test")
    admin_page.get_by_label("Placeholder:").fill("Please select something....")
    admin_page.get_by_label("Html id:").fill("test")
    admin_page.get_by_role("button", name="Save", exact=True).click()
    admin_page.get_by_role("link", name="Form kit schemas").click()
    admin_page.get_by_role("link", name="Add form kit schema").click()
    admin_page.get_by_label("Label:").fill("Test Schema")
    admin_page.get_by_role("button", name="Save and continue editing").click()
    admin_page.get_by_role("link", name="Add another Form components").click()
    admin_page.get_by_role("button", name="Save", exact=True).click()
    admin_page.get_by_role("link", name="Form componentss").click()
    admin_page.get_by_role("link", name="Add form components").click()
    admin_page.get_by_label("Schema:").select_option(index=0)
    admin_page.get_by_label("Node:").select_option(index=0)
    admin_page.get_by_label("Label:").fill("'test' node in 'test' schema")
    admin_page.get_by_role("button", name="Save", exact=True).click()


# Tests that the admin pages from importing SF 1.1 make sense


@pytest.mark.django_db()
def test_import_sf11(SF_1_1):
    """
    This tests that we can successfully import the 'SF11' form from Partisipa
    """
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    schema = BaseModel.model_validate([SF_1_1])
    schema_in_the_db = models.FormKitSchema.from_pydantic(schema)

    # Check that schema sections retained their order
    # There is a top level node
    assert schema_in_the_db.nodes.first().label == "SF_1_1"

    # The 'top level' node has the same child nodes as the input schema
    child_nodes = schema_in_the_db.nodes.first().get_node(recursive=True).children

    assert [n["id"] for n in SF_1_1["children"]] == [n.id for n in child_nodes]

    # The inputs and outputs
    partA_in = SF_1_1["children"][0]["children"]
    partA_schema = child_nodes[0].children
    partA_out = [n.model_dump() for n in partA_schema]

    # Nested (children) should retain their order
    assert (
        [n["key"] for n in partA_in]
        == [n.key for n in partA_schema]
        == [n["key"] for n in partA_out]
    )
    assert (
        [n["label"] for n in partA_in]
        == [n.label for n in partA_schema]
        == [n["label"] for n in partA_out]
    )

    assert partA_in[0]["key"] == partA_schema[0].key
    assert partA_in[0]["name"] == partA_schema[0].name
    assert partA_in[0]["label"] == partA_schema[0].label
    assert partA_in[0]["placeholder"] == partA_schema[0].placeholder
    assert partA_in[0]["options"] == partA_schema[0].options

    assert partA_in[0]["key"] == partA_out[0]["key"]
    assert partA_in[0]["name"] == partA_out[0]["name"]
    assert partA_in[0]["label"] == partA_out[0]["label"]
    assert partA_in[0]["placeholder"] == partA_out[0]["placeholder"]
    assert partA_in[0]["options"] == partA_out[0]["options"]


@playwright
@pytest.mark.django_db()
def test_admin_actions_sf11(SF_1_1, admin_page: Page):
    """
    This tests that we can successfully import the 'SF11' form from Partisipa
    """
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    schema_json = [SF_1_1]
    schema = BaseModel.model_validate(schema_json)
    models.FormKitSchema.from_pydantic(schema)
    admin_page.get_by_role("link", name="Form kit schema nodes").click()


@playwright
@pytest.mark.django_db()
def test_import_1321(TF_13_2_1, admin_page):
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    schema_json = [TF_13_2_1]
    models.FormKitSchema.from_pydantic(BaseModel.model_validate(schema_json))
    admin_page.get_by_role("link", name="Form kit schema nodes").click()
    admin_page.get_by_role("link", name="repeaterProjectProgress").click()


@playwright
@pytest.mark.django_db
@pytest.mark.parametrize(
    "schema",
    [
        "CFM_12_FF_12",
        "CFM_2_FF_4",
        "FF_14",
        "POM_1",
        "SF_1_1",
        "SF_1_2",
        "SF_1_3",
        "SF_2_3",
        "SF_4_1",
        "SF_4_2",
        "SF_6_2",
        "TF_13_2_1",
        "TF_6_1_1",
    ],
)
@pytest.mark.django_db()
def test_admin_all_forms(admin_page, schema):
    from formkit_ninja.schemas import Schemas

    schemas = Schemas()
    schema_json = schemas.as_json(schema)
    from formkit_ninja.formkit_schema import FormKitSchema as BaseModel

    schema_as_pydantic = BaseModel.model_validate(schema_json)

    _ = models.FormKitSchema.from_pydantic(BaseModel.model_validate(schema_as_pydantic))
    admin_page.pause()



# Generated by Django 4.2.5 on 2023-10-03 00:08

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0019_formkitschemanode_track_change_and_more"),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name="formcomponents",
            name="order_on_update_option",
        ),
        pgtrigger.migrations.RemoveTrigger(
            model_name="nodechildren",
            name="order_on_update_option",
        ),
        pgtrigger.migrations.RemoveTrigger(
            model_name="option",
            name="order_on_update_option",
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formcomponents",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_update_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (pg_trigger_depth() = 0)",
                    func='\n                -- Do not allow a "null" value\n                -- This stops Django from dumbly updating\n                -- which can break the trigger\n                if NEW."order" IS NULL then\n                    NEW."order" = OLD."order";\n                end if;\n                if NEW."order" > OLD."order" then\n                    update formkit_ninja_formcomponents\n                    set "order" = "order"- 1\n                    where "order" <= NEW."order"\n                    and "order" > OLD."order"\n                    and "schema_id" = NEW."schema_id"\n                    and "id" <> NEW."id";\n                else\n                    update formkit_ninja_formcomponents\n                    set "order" = "order"+ 1\n                    where "order" >= NEW."order"\n                    and "order" < OLD."order"\n                    and "schema_id" = NEW."schema_id"\n                    and "id" <> NEW."id";\n                end if;\n                RETURN NEW;\n        ',
                    hash="432443481a60d58c40ceb3204bda0e3ad58ce9aa",
                    operation="UPDATE",
                    pgid="pgtrigger_order_on_update_option_a5326",
                    table="formkit_ninja_formcomponents",
                    when="AFTER",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="nodechildren",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_update_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (pg_trigger_depth() = 0)",
                    func='\n                -- Do not allow a "null" value\n                -- This stops Django from dumbly updating\n                -- which can break the trigger\n                if NEW."order" IS NULL then\n                    NEW."order" = OLD."order";\n                end if;\n                if NEW."order" > OLD."order" then\n                    update formkit_ninja_nodechildren\n                    set "order" = "order"- 1\n                    where "order" <= NEW."order"\n                    and "order" > OLD."order"\n                    and "parent_id" = NEW."parent_id"\n                    and "id" <> NEW."id";\n                else\n                    update formkit_ninja_nodechildren\n                    set "order" = "order"+ 1\n                    where "order" >= NEW."order"\n                    and "order" < OLD."order"\n                    and "parent_id" = NEW."parent_id"\n                    and "id" <> NEW."id";\n                end if;\n                RETURN NEW;\n        ',
                    hash="f8eaa7c49b5c81e7b0fa3830557746f40e2498bd",
                    operation="UPDATE",
                    pgid="pgtrigger_order_on_update_option_44228",
                    table="formkit_ninja_nodechildren",
                    when="AFTER",
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="option",
            trigger=pgtrigger.compiler.Trigger(
                name="order_on_update_option",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition="WHEN (pg_trigger_depth() = 0)",
                    func='\n                -- Do not allow a "null" value\n                -- This stops Django from dumbly updating\n                -- which can break the trigger\n                if NEW."order" IS NULL then\n                    NEW."order" = OLD."order";\n                end if;\n                if NEW."order" > OLD."order" then\n                    update formkit_ninja_option\n                    set "order" = "order"- 1\n                    where "order" <= NEW."order"\n                    and "order" > OLD."order"\n                    and "group_id" = NEW."group_id"\n                    and "id" <> NEW."id";\n                else\n                    update formkit_ninja_option\n                    set "order" = "order"+ 1\n                    where "order" >= NEW."order"\n                    and "order" < OLD."order"\n                    and "group_id" = NEW."group_id"\n                    and "id" <> NEW."id";\n                end if;\n                RETURN NEW;\n        ',
                    hash="3f44c5db88a1344000e61868009e4b424d7c29ef",
                    operation="UPDATE",
                    pgid="pgtrigger_order_on_update_option_417cb",
                    table="formkit_ninja_option",
                    when="AFTER",
                ),
            ),
        ),
    ]



# Generated by Django 4.2.2 on 2023-08-29 08:29

import uuid

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("formkit_ninja", "0009_formkitschemanode_text_content_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="option",
            options={"ordering": ("group", "order")},
        ),
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="children",
        ),
        migrations.AlterField(
            model_name="formcomponents",
            name="id",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                primary_key=True,
                serialize=False,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="formkitschema",
            name="id",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                primary_key=True,
                serialize=False,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="formkitschemanode",
            name="id",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                primary_key=True,
                serialize=False,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="membership",
            name="id",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                primary_key=True,
                serialize=False,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="option",
            name="group",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.optiongroup",
            ),
        ),
        migrations.AlterField(
            model_name="option",
            name="id",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                primary_key=True,
                serialize=False,
                unique=True,
            ),
        ),
        migrations.AlterField(
            model_name="optiongroup",
            name="content_type",
            field=models.ForeignKey(
                blank=True,
                help_text="This is an optional reference to the original source object for this set of options (typically a table from which we copy options)",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="contenttypes.contenttype",
            ),
        ),
        migrations.DeleteModel(
            name="NodeChildren",
        ),
    ]



# Generated by Django 4.2.4 on 2023-09-04 10:05

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0011_nodechildren_formkitschemanode_children"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="formkitschemanode",
            name="group",
        ),
        migrations.RemoveField(
            model_name="option",
            name="field",
        ),
        migrations.AlterField(
            model_name="formkitschema",
            name="label",
            field=models.CharField(
                blank=True,
                help_text="Used as a human-readable label",
                max_length=1024,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="nodechildren",
            name="child",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.AlterField(
            model_name="nodechildren",
            name="parent",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="parent",
                to="formkit_ninja.formkitschemanode",
            ),
        ),
        migrations.DeleteModel(
            name="Membership",
        ),
    ]



# Generated by Django 4.2.5 on 2023-10-01 05:33

from django.db import migrations

from formkit_ninja import triggers


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0016_alter_nodechildren_options_and_more"),
    ]

    operations = [
        # Create the SQL sequence we'll use for nodechildren
        migrations.RunSQL(
            *triggers.create_sequence_migration(triggers.NODE_CHILDREN_CHANGE_ID)
        ),
    ]



import pytest

from formkit_ninja.formkit_schema import (DiscriminatedNodeType,
                                          FormKitSchemaDOMNode, TextNode)
from formkit_ninja.models import FormKitSchemaNode


@pytest.mark.django_db
def test_element():
    """
    An "el" (div) in the database
    should return a DomNode
    """
    node_in_db = FormKitSchemaNode.objects.create(
        node_type="$el", node={"$el": "div"}, label="A node"
    )
    node_from_db = node_in_db.to_pydantic()
    assert isinstance(node_from_db.root, FormKitSchemaDOMNode)


def test_validate():
    assert isinstance(
        DiscriminatedNodeType.model_validate({"$el": "div"}).root, FormKitSchemaDOMNode
    )

    assert isinstance(
        DiscriminatedNodeType.model_validate({"$formkit": "text"}).root, TextNode
    )



import json
import pathlib
from importlib.resources import files

from formkit_ninja import schemas as schema_path
from formkit_ninja.formkit_schema import FormKitNode


class Schemas:
    def __init__(self):
        # All json files
        schema_files: list[pathlib.Path] = list(files(schema_path).glob("*.json"))
        # Name of the schema: path
        schemas = {path_.name[:-5]: path_ for path_ in schema_files}
        self.schemas = schemas

    def list_schemas(self) -> list[str]:
        return self.schemas.keys()

    def as_text(self, schema: str):
        return self.schemas[schema].read_text()

    def as_json(self, schema: str) -> dict:
        return json.loads(self.as_text(schema))

    def as_dict(self, schema: str):
        return json.dumps(self.as_json(schema))

    def import_all(self):
        from formkit_ninja.models import FormKitSchemaNode

        for schema in self.schemas.keys():
            node: FormKitNode = FormKitNode.parse_obj(self.as_json(schema))
            parsed_node = node.root
            list(FormKitSchemaNode.from_pydantic(parsed_node))[0]



import os
from textwrap import dedent

from jinja2 import Environment, PackageLoader, Template, select_autoescape

from formkit_ninja.parser.type_convert import NodePath


def strip_empty_lines(text: str):
    return os.linesep.join([s for s in text.splitlines() if s])


def get_env():
    return Environment(
        loader=PackageLoader("formkit_ninja.parser"),
        autoescape=select_autoescape(),
        trim_blocks=True,
        lstrip_blocks=True,
    )


def test_nested_group_node(nested_group_node: NodePath):
    assert nested_group_node.is_repeater is False
    assert nested_group_node.classname == "Bar"
    assert nested_group_node.classname_schema == "BarSchema"
    assert nested_group_node.django_type == "OneToOneField"
    assert nested_group_node.is_group is True
    assert nested_group_node.is_child is False


def test_number_node_field(number_node: NodePath):
    assert number_node.is_repeater is False


def test_pd_number_node_field(pydantic_class_template: Template, number_node: NodePath):
    text = pydantic_class_template.render(this=number_node)
    expect = "    foonum: int\n"
    assert text.strip() == dedent(expect).strip()


def test_pd_group_node_field(pydantic_class_template: Template, group_node: NodePath):
    text = pydantic_class_template.render(this=group_node)
    expect = "class Foo(BaseModel):\n    foonum: int | None = None\n"
    assert text.strip() == dedent(expect).strip()


def test_pd_nested_group_node_field(
    pydantic_class_template: Template, nested_group_node: NodePath
):
    text = pydantic_class_template.render(this=nested_group_node)
    expect = """
        class BarFoo(BaseModel):
            foonum: int | None = None
        class Bar(BaseModel):
            foo: BarFoo | None = None
        """
    assert text.strip() == dedent(expect).strip()


def test_group_node_field(django_class_template: Template, group_node: NodePath):
    text = django_class_template.render(this=group_node)
    expect = "class Foo(models.Model):\n    foonum = models.IntegerField(null=True, blank=True)\n"
    assert text.strip() == dedent(expect).strip()


def test_nested_group_node_field(
    django_class_template: Template, nested_group_node: NodePath
):
    text = django_class_template.render(this=nested_group_node)
    expect = """
        class BarFoo(models.Model):
            foonum = models.IntegerField(null=True, blank=True)
        class Bar(models.Model):
            foo = models.OneToOneField(BarFoo, on_delete=models.CASCADE)
        """
    assert text.strip() == dedent(expect).strip()


def test_admin_group_node_field(admin_template: Template, group_node: NodePath):
    text = admin_template.render(this=group_node)
    expect = """
        @admin.register(models.Foo)
        class FooAdmin(admin.ModelAdmin):
            list_display = [
                "foonum",
            ]
            readonly_fields = [
                "foonum",
            ]
    """
    assert text.strip() == dedent(expect).strip()


def test_admin_py_group_node_field(admin_py_template: Template, group_node: NodePath):
    text = admin_py_template.render(nodepaths=[group_node])
    expect = '''
"""
Don't make changes to this code directly
Instead, make changes to the template and re-generate this file
"""

from django.contrib import admin
from . import models

class ReadOnlyInline(admin.TabularInline):
    def has_change_permission(self, request, obj=None):
        return False
    def has_add_permission(self, request, obj=None):
        return False
    def has_delete_permission(self, request, obj=None):
        return False

@admin.register(models.Foo)
class FooAdmin(admin.ModelAdmin):
    list_display = [
        "foonum",
    ]
    readonly_fields = [
        "foonum",
    ]
    '''
    assert text.strip() == dedent(expect).strip()


def test_admin_nested_group_node_field(
    admin_template: Template, nested_group_node: NodePath
):
    text = admin_template.render(this=nested_group_node)
    expect = """
        @admin.register(models.BarFoo)
        class BarFooAdmin(admin.ModelAdmin):
            list_display = [
                "foonum",
            ]
            readonly_fields = [
                "foonum",
            ]
        @admin.register(models.Bar)
        class BarAdmin(admin.ModelAdmin):
            list_display = [
                "foo",
            ]
            readonly_fields = [
                "foo",
            ]
        """
    assert text.strip() == dedent(expect).strip()


def test_api_nested_group_node_field(
    api_template: Template, nested_group_node: NodePath
):
    text = api_template.render(this=nested_group_node)
    expect = """
        @router.get("barfoo", response=list[schema_out.BarFooSchema], exclude_none=True)
        def barfoo(request):
            queryset = models.BarFoo.objects.all()
            return queryset
        @router.get("bar", response=list[schema_out.BarSchema], exclude_none=True)
        def bar(request):
            queryset = models.Bar.objects.all()
            queryset = queryset.select_related(
                "foo",
            )
            return queryset
        """
    assert text.strip() == dedent(expect).strip()


def test_api_nested_repeater_node_field(
    api_template: Template, nested_repeater_node: NodePath
):
    text = api_template.render(this=nested_repeater_node)
    expect = """
        @router.get("barfoo", response=list[schema_out.BarFooSchema], exclude_none=True)
        def barfoo(request):
            queryset = models.BarFoo.objects.all()
            return queryset
        @router.get("bar", response=list[schema_out.BarSchema], exclude_none=True)
        def bar(request):
            queryset = models.Bar.objects.all()
            queryset = queryset.prefetch_related(
                "foo",
            )
            return queryset
        """
    assert text.strip() == dedent(expect).strip()


def test_schema_out_nested_group_node_field(
    schema_out_template: Template, nested_group_node: NodePath
):
    text = schema_out_template.render(this=nested_group_node)
    expect = """
        class BarFooSchema(Schema):
            foonum: int | None = None
        class BarSchema(Schema):
            foo: BarFoo | None = None
    """
    assert text.strip() == dedent(expect).strip()



# Generated by Django 4.2.6 on 2023-10-16 14:52

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0021_formkitschemanode_is_active"),
    ]

    operations = [
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="soft_delete",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='UPDATE "formkit_ninja_formkitschemanode" SET is_active = False WHERE "id" = OLD."id"; RETURN NULL;',
                    hash="7e81f11bf057637a44641b9885a497624546c502",
                    operation="DELETE",
                    pgid="pgtrigger_soft_delete_69d8d",
                    table="formkit_ninja_formkitschemanode",
                    when="BEFORE",
                ),
            ),
        ),
    ]



from __future__ import annotations

import itertools
import logging
import uuid
import warnings
from keyword import iskeyword, issoftkeyword
from typing import Iterable, TypedDict, get_args

import pghistory
import pgtrigger
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.aggregates import ArrayAgg
from django.db import models, transaction
from django.db.models import F, Q
from django.db.models.aggregates import Max
from django.db.models.functions import Greatest
from rich.console import Console

from formkit_ninja import formkit_schema, triggers

console = Console()
log = console.log

logger = logging.getLogger()


def check_valid_django_id(key: str):
    if not key.isidentifier() or iskeyword(key) or issoftkeyword(key):
        raise TypeError(
            f"{key} cannot be used as a keyword. Should be a valid python identifier"
        )
    if key[0].isdigit():
        raise TypeError(f"{key} is not valid, it cannot start with a digit")
    if key[-1] == "_":
        raise TypeError(f"{key} is not valid, it cannot end with an underscore")


class UuidIdModel(models.Model):
    """
    Consistently use fields which will
    help with syncing data:
     - UUID field is the ID
     - Created field
     - Last Modified field
     - updated_by (optional)
     - created_by (optional)
    """

    class Meta:
        abstract = True

    id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, editable=False, unique=True
    )
    created = models.DateTimeField(auto_now_add=True, blank=True, null=True)
    updated = models.DateTimeField(auto_now=True, blank=True, null=True)
    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="+",
        blank=True,
        null=True,
    )
    updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.PROTECT,
        related_name="+",
        blank=True,
        null=True,
    )


class OptionDict(TypedDict):
    value: str
    label: str


class OptionGroup(models.Model):
    """
    This intended to be a "collection" of choices
    For instance all the values in a single PNDS zTable
    Also intended to allow users to add / modify their __own__ 'Options'
    for idb and formkit to recognize
    """

    group = models.CharField(
        max_length=1024,
        primary_key=True,
        help_text="The label to use for these options",
    )
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="This is an optional reference to the original source object for this set of options (typically a table from which we copy options)",
    )

    # If the object is a "Content Type" we expect it to have a similar layout to this

    def save(self, *args, **kwargs):
        # Prior to save ensure that content_type, if present, fits suitable schema
        if self.content_type:
            klass = self.content_type.model_class()
            try:
                if klass._meta.get_field("value") is None or not hasattr(
                    klass, "label_set"
                ):
                    raise ValueError(
                        f"Expected {klass} to have a 'value' field and a 'label_set' attribute"
                    )
            except Exception as E:
                raise ValueError(
                    f"Expected {klass} to have a 'value' field and a 'label_set' attribute"
                ) from E
        return super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.group}"

    @classmethod
    def copy_table(
        cls,
        model: models.Model,
        field: str,
        language: str | None = "en",
        group: str | None = None,
    ):
        """
        Copy an existing table of options into this OptionGroup
        """

        with transaction.atomic():
            group, group_created = cls.objects.get_or_create(
                group=group, content_type=ContentType.objects.get_for_model(model)
            )
            log(group)

            for obj in model.objects.values("pk", field):
                option, option_created = Option.objects.get_or_create(
                    object_id=obj["pk"],
                    group=group,
                    value=obj["pk"],
                )
                OptionLabel.objects.get_or_create(
                    option=option, label=obj[field] or "", lang=language
                )


class OptionQuerySet(models.Manager):
    """
    Prefetched "labels" for performance
    """

    def get_queryset(self):
        """
        Added a prefetch_related to the queryset
        """
        lang_codes = (n[0] for n in settings.LANGUAGES)

        label_model = OptionLabel
        annotated_fields = {
            f"label_{lang}": label_model.objects.filter(
                lang=lang, option=models.OuterRef("pk")
            )
            for lang in lang_codes
        }
        annotated_fields_subquery = {
            field: models.Subquery(
                query.values("label")[:1], output_field=models.CharField()
            )
            for field, query in annotated_fields.items()
        }
        return super().get_queryset().annotate(**annotated_fields_subquery)


class Option(UuidIdModel):
    """
    This is a key/value field representing one "option" for a FormKit property
    The translated values for this option are in the `Translatable` table
    """

    object_id = models.IntegerField(
        null=True,
        blank=True,
        help_text="This is a reference to the primary key of the original source object (typically a PNDS ztable ID) or a user-specified ID for a new group",
    )
    last_updated = models.DateTimeField(auto_now=True)
    group = models.ForeignKey(
        OptionGroup, on_delete=models.CASCADE, null=True, blank=True
    )
    # is_active = models.BooleanField(default=True)
    order = models.IntegerField(null=True, blank=True)

    class Meta:
        triggers = triggers.update_or_insert_group_trigger("group_id")
        constraints = [
            models.UniqueConstraint(
                fields=["group", "object_id"], name="unique_option_id"
            )
        ]
        ordering = (
            "group",
            "order",
        )

    value = models.CharField(max_length=1024)
    order_with_respect_to = "group"

    objects = OptionQuerySet()

    @classmethod
    def from_pydantic(
        cls,
        options: list[str] | list[OptionDict],
        group: OptionGroup | None = None,
    ) -> Iterable["Option"]:
        """
        Yields "Options" in the database based on the input given
        """
        for option in options:
            if isinstance(option, str):
                opt = cls(value=option, group=group)
                # Capture the effects of triggers
                # Else we override with the 'default' value of 0
                opt.save()
                opt.refresh_from_db()
                OptionLabel.objects.create(option=opt, lang="en", label=option)
            elif isinstance(option, dict) and option.keys() == {"value", "label"}:
                opt = cls(value=option["value"], group=group)
                OptionLabel.objects.create(option=opt, lang="en", label=option["label"])
            else:
                console.log(f"[red]Could not format the given object {option}")
                continue
            yield opt

    def __str__(self):
        if self.group:
            return f"{self.group.group}::{self.value}"
        else:
            return f"No group: {self.value}"


class OptionLabel(models.Model):
    option = models.ForeignKey("Option", on_delete=models.CASCADE)
    label = models.CharField(max_length=1024)
    lang = models.CharField(
        max_length=4,
        default="en",
        choices=(("en", "English"), ("tet", "Tetum"), ("pt", "Portugese")),
    )

    def save(self, *args, **kwargs):
        """
        When saved, save also my "option" so that its last_updated is set
        """
        if self.option is not None:
            self.option.save()
        return super().save(*args, **kwargs)

    class Meta:
        constraints = [
            models.UniqueConstraint(
                fields=["option", "lang"], name="unique_option_label"
            )
        ]


class FormComponents(UuidIdModel):
    """
    A model relating "nodes" of a schema to a schema with model ordering
    """

    schema = models.ForeignKey("FormKitSchema", on_delete=models.CASCADE)
    # This is null=True so that a new FormComponent can be added from the admin inline
    node = models.ForeignKey(
        "FormKitSchemaNode", on_delete=models.CASCADE, null=True, blank=True
    )
    label = models.CharField(
        max_length=1024,
        help_text="Used as a human-readable label",
        null=True,
        blank=True,
    )
    order = models.IntegerField(null=True, blank=True)
    order_with_respect_to = "schema"

    class Meta:
        triggers = triggers.update_or_insert_group_trigger("schema_id")
        ordering = ("schema", "order")

    def __str__(self):
        return f"{self.node}[{self.order}]: {self.schema}"


class NodeChildrenManager(models.Manager):
    """
    Adds aggregation and filtering for client side data
    of NodeChildren relations
    """

    def aggregate_changes_table(self, latest_change: int | None = None):
        values = (
            self.get_queryset()
            .values("parent_id")
            .annotate(
                children=ArrayAgg("child", ordering=F("order")),
            )
            .annotate(Max("child__track_change"))
            .annotate(
                latest_change=Greatest(
                    "child__track_change__max", "parent__track_change"
                )
            )
        )
        if latest_change:
            values = values.filter(
                Q(latest_change__gt=latest_change)
                | Q(parent__latest_change__gt=latest_change)
            )
        return values.values_list("parent_id", "latest_change", "children", named=True)


class NodeChildren(models.Model):
    """
    This is an ordered m2m model representing
    the "children" of an HTML element
    """

    parent = models.ForeignKey(
        "FormKitSchemaNode",
        on_delete=models.CASCADE,
        related_name="parent",
    )
    child = models.ForeignKey("FormKitSchemaNode", on_delete=models.CASCADE)
    order = models.IntegerField(null=True, blank=True)
    track_change = models.BigIntegerField(null=True, blank=True)
    order_with_respect_to = "parent"

    class Meta:
        triggers = [
            *triggers.update_or_insert_group_trigger("parent_id"),
            triggers.bump_sequence_value(
                sequence_name=triggers.NODE_CHILDREN_CHANGE_ID
            ),
        ]
        ordering = (
            "parent_id",
            "order",
        )

    objects = NodeChildrenManager()


class NodeQS(models.QuerySet):
    def from_change(self, track_change: int = -1):
        return self.filter(track_change__gt=track_change)

    def to_response(
        self, ignore_errors: bool = True, options: bool = True
    ) -> Iterable[tuple[uuid.UUID, int, formkit_schema.Node | str | None, bool]]:
        """
        Return a set of FormKit nodes
        """
        node: FormKitSchemaNode
        for node in self.all():
            try:
                if node.is_active:
                    yield (
                        node.id,
                        node.track_change,
                        node.get_node(recursive=False, options=options),
                        node.protected,
                    )
                else:
                    yield node.id, node.track_change, None, node.protected
            except Exception as E:
                if not ignore_errors:
                    raise
                warnings.warn(f"An unparseable FormKit node was hit at {node.pk}")
                warnings.warn(f"{E}")


@pghistory.track()
@pgtrigger.register(
    pgtrigger.Protect(
        # If the node is protected, delete is not allowed
        name="protect_node_deletes_and_updates",
        operation=pgtrigger.Delete,
        condition=pgtrigger.Q(old__protected=True),
    ),
    pgtrigger.Protect(
        # If both new and old values are "protected", updates are not allowed
        name="protect_node_updates",
        operation=pgtrigger.Update,
        condition=pgtrigger.Q(old__protected=True) & pgtrigger.Q(new__protected=True),
    ),
    pgtrigger.SoftDelete(name="soft_delete", field="is_active"),
    triggers.bump_sequence_value("track_change", triggers.NODE_CHANGE_ID),
)
class FormKitSchemaNode(UuidIdModel):
    """
    This represents a single "Node" in a FormKit schema.
    There are several different types of node which may be defined:
    FormKitSchemaDOMNode
    | FormKitSchemaComponent
    | FormKitSchemaTextNode
    | FormKitSchemaCondition
    | FormKitSchemaFormKit
    """

    objects = NodeQS.as_manager()

    NODE_TYPE_CHOICES = (
        ("$cmp", "Component"),  # Not yet implemented
        ("text", "Text"),
        ("condition", "Condition"),  # Not yet implemented
        ("$formkit", "FormKit"),
        ("$el", "Element"),
        ("raw", "Raw JSON"),  # Not yet implemented
    )
    FORMKIT_CHOICES = [(t, t) for t in get_args(formkit_schema.FORMKIT_TYPE)]

    ELEMENT_TYPE_CHOICES = [("p", "p"), ("h1", "h1"), ("h2", "h2"), ("span", "span")]
    node_type = models.CharField(
        max_length=256, choices=NODE_TYPE_CHOICES, blank=True, help_text=""
    )
    description = models.CharField(
        max_length=4000,
        null=True,
        blank=True,
        help_text="Decribe the type of data / reason for this component",
    )
    label = models.CharField(
        max_length=1024,
        help_text="Used as a human-readable label",
        null=True,
        blank=True,
    )
    option_group = models.ForeignKey(
        OptionGroup, null=True, blank=True, on_delete=models.PROTECT
    )
    children = models.ManyToManyField(
        "self", through=NodeChildren, symmetrical=False, blank=True
    )
    is_active = models.BooleanField(default=True)
    protected = models.BooleanField(default=False)

    node = models.JSONField(
        null=True,
        blank=True,
        help_text="A JSON representation of select parts of the FormKit schema",
    )

    additional_props = models.JSONField(
        null=True,
        blank=True,
        help_text="User space for additional, less used props",
    )

    text_content = models.TextField(
        null=True,
        blank=True,
        help_text="Content for a text element, for children of an $el type component",
    )
    track_change = models.BigIntegerField(null=True, blank=True)

    def __str__(self):
        return f"Node: {self.label}" if self.label else f"{self.node_type} {self.id}"

    def save(self, *args, **kwargs):
        """
        On save validate the 'node' field matches the 'FormKitNode'
        """
        # rename `formkit` to `$formkit`
        if isinstance(self.node, dict) and "formkit" in self.node:
            self.node.update({"$formkit": self.node.pop("formkit")})
        # We're also going to verify that the 'key' is a valid identifier
        # Keep in mind that the `key` may be used as part of a model so
        # should be valid Django fieldname too
        if isinstance(self.node, dict) and "name" in self.node:
            key: str = self.node.get("name", None)
            check_valid_django_id(key)
        return super().save(*args, **kwargs)

    @property
    def node_options(self) -> str | list[dict] | None:
        """
        Because "options" are translated and
        separately stored, this step is necessary to
        reinstate them
        """
        if isinstance(self.node, dict) and "options" in self.node:
            if opts := self.node.get("options"):
                return opts

        if not self.option_group:
            return None
        options = self.option_group.option_set.all().prefetch_related("optionlabel_set")
        # options: Iterable[Option] = self.option_set.all().prefetch_related("optionlabel_set")
        # TODO: This is horribly slow
        return [
            {"value": option.value, "label": f"{option.optionlabel_set.first().label}"}
            for option in options
        ]

    def get_node_values(
        self, recursive: bool = True, options: bool = True
    ) -> str | dict:
        """
        Reify a 'dict' instance suitable for creating
        a FormKit Schema node from
        """
        # Text element
        if not self.node:
            if self.text_content:
                return self.text_content
            return {}
        values = {**self.node}

        # Options may come from a string in the node, or
        # may come from an m2m
        if options and self.node_options:
            values["options"] = self.node_options
        if recursive:
            children = [
                c.get_node_values()
                for c in self.children.order_by("nodechildren__order")
            ]
            if children:
                values["children"] = children
        if self.additional_props and len(self.additional_props) > 0:
            values["additional_props"] = self.additional_props

        if values == {}:
            if self.node_type == "$el":
                values.update({"$el": "span"})
            elif self.node_type == "$formkit":
                values.update({"$formkit": "text"})

        return values

    def get_node(
        self, recursive=False, options=False, **kwargs
    ) -> formkit_schema.Node | str:
        """
        Return a "decorated" node instance
        with restored options and translated fields
        """
        if self.text_content or self.node_type == "text":
            return self.text_content or ""
        if self.node == {} or self.node is None:
            if self.node_type == "$el":
                node_content = {"$el": "span"}
            elif self.node_type == "$formkit":
                node_content = {"$formkit": "text"}
        else:
            node_content = self.get_node_values(
                **kwargs, recursive=recursive, options=options
            )

        formkit_node = formkit_schema.DiscriminatedNodeType.model_validate(node_content)
        return formkit_node.root

    @classmethod
    def _from_props_instance(cls, input_model: formkit_schema.FormKitSchemaProps):
        instance = cls()
        log(f"[green]Creating {instance}")
        for label_field in ("name", "id", "key", "label"):
            if label := getattr(input_model, label_field, None):
                instance.label = label
                break

        # Node types
        if props := getattr(input_model, "additional_props", None):
            instance.additional_props = props
        try:
            node_type = getattr(input_model, "node_type")
        except Exception as E:
            raise E
        if node_type == "condition":
            instance.node_type = "condition"
        elif node_type == "formkit":
            instance.node_type = "$formkit"
        elif node_type == "element":
            instance.node_type = "$el"
        elif node_type == "component":
            instance.node_type = "$cmp"

        log(f"[green]Yielding: {instance}")

        # Must save the instance before  adding "options" or "children"
        instance.node = input_model.model_dump(
            exclude={
                "options",
                "children",
                "additional_props",
                "node_type",
            },
            exclude_none=True,
            exclude_unset=True,
            by_alias=True,
        )
        # Where an alias is used ("el", ) restore it to the expected value
        # of a FormKit schema node
        for pydantic_key, db_key in (("el", "$el"), ("formkit", "$formkit")):
            if db_value := instance.node.pop(pydantic_key, None):
                instance.node[db_key] = db_value

        instance.save()
        # Add the "options" if it is a 'text' type getter
        options: formkit_schema.OptionsType = getattr(input_model, "options", None)

        if isinstance(options, str):
            # Maintain this as it is probably a `$get...` options call
            # to a Javascript function
            instance.node["options"] = options
            instance.save()

        elif isinstance(options, Iterable):
            # Create a new "group" to assign these options to
            # Here we use a random UUID as the group name
            instance.option_group = OptionGroup.objects.create(
                group=f"Auto generated group for {str(instance)} {uuid.uuid4().hex[0:8]}"
            )
            for option in Option.from_pydantic(options, group=instance.option_group):
                pass
            instance.save()
        # Retain input strings without splitting to a list
        children = getattr(input_model, "children", []) or []
        if isinstance(children, str):
            child_node = next(iter(cls.from_pydantic(children)))
            console.log(f"    {child_node}")
            instance.children.add(child_node)
        elif isinstance(children, Iterable):
            for c_n in children:
                child_node = next(iter(cls.from_pydantic(c_n)))
                console.log(f"    {child_node}")
                instance.children.add(child_node)

        yield instance

    @classmethod
    def from_pydantic(
        cls,
        input_models: (
            formkit_schema.FormKitSchemaProps
            | Iterable[formkit_schema.FormKitSchemaProps]
            | str
        ),
    ):
        if isinstance(input_models, str):
            yield cls.objects.create(
                node_type="text", label=input_models, text_content=input_models
            )

        if isinstance(input_models, formkit_schema.DiscriminatedNodeType):
            yield from cls.from_pydantic(input_models.root)

        elif isinstance(input_models, formkit_schema.DiscriminatedNodeTypeSchema):
            for node in input_models.root:
                yield from cls.from_pydantic(node)

        elif isinstance(input_models, Iterable) and not isinstance(
            input_models, formkit_schema.FormKitSchemaProps
        ):
            yield from (cls.from_pydantic(n) for n in input_models)

        elif isinstance(input_models, formkit_schema.FormKitSchemaProps):
            yield from cls._from_props_instance(input_models)

        else:
            raise TypeError(
                f"Expected FormKitNode or Iterable[FormKitNode], got {type(input_models)}"
            )

    def to_pydantic(self, recursive=False, options=False, **kwargs):
        if self.text_content:
            return self.text_content
        return formkit_schema.DiscriminatedNodeType.model_validate(
            self.get_node_values(recursive=recursive, options=options, **kwargs)
        )


class SchemaManager(models.Manager):
    """
    Provides prefetching which we'll almost always want to have
    """

    def get_queryset(self):
        return super().get_queryset().prefetch_related("nodes", "nodes__children")


class FormKitSchema(UuidIdModel):
    """
    This represents a "FormKitSchema" which is an heterogenous
    collection of items.
    """

    label = models.CharField(
        max_length=1024,
        null=True,
        blank=True,
        help_text="Used as a human-readable label",
        unique=True,
        default=uuid.uuid4,
    )
    nodes = models.ManyToManyField(FormKitSchemaNode, through=FormComponents)
    objects = SchemaManager()

    def get_schema_values(self, recursive=False, options=False, **kwargs):
        """
        Return a list of "node" dicts
        """
        nodes: Iterable[FormKitSchemaNode] = self.nodes.order_by(
            "formcomponents__order"
        )
        for node in nodes:
            yield node.get_node_values(recursive=recursive, options=options, **kwargs)

    def to_pydantic(self):
        values = list(self.get_schema_values())
        return formkit_schema.FormKitSchema.parse_obj(values)

    def __str__(self):
        return f"{self.label}" or f"{str(self.id)[:8]}"

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    @classmethod
    def from_pydantic(
        cls, input_model: formkit_schema.FormKitSchema, label: str | None = None
    ) -> "FormKitSchema":
        """
        Converts a given Pydantic representation of a Schema
        to Django database fields
        """
        instance = cls.objects.create(label=label)
        for node in itertools.chain.from_iterable(
            FormKitSchemaNode.from_pydantic(input_model.root)
        ):
            log(f"[yellow]Saving {node}")
            node.save()
            FormComponents.objects.create(
                schema=instance, node=node, label=str(f"{str(instance)} {str(node)}")
            )
        logger.info("Schema load from JSON done")
        return instance

    @classmethod
    def from_json(cls, input_file: dict):
        """
        Converts a given JSON string to a suitable
        Django representation
        """
        schema_instance = formkit_schema.FormKitSchema.parse_obj(input_file)
        return cls.from_pydantic(schema_instance)


class SchemaLabel(models.Model):
    """
    This intended to hold translations of Partisipa schema definitions.
    The title.
    """

    schema = models.ForeignKey("FormKitSchema", on_delete=models.CASCADE)
    label = models.CharField(max_length=1024)
    lang = models.CharField(
        max_length=4,
        default="en",
        choices=(("en", "English"), ("tet", "Tetum"), ("pt", "Portugese")),
    )


class SchemaDescription(models.Model):
    """
    This intended to hold translations of Partisipa schema definitions.
    The description.
    """

    schema = models.ForeignKey("FormKitSchema", on_delete=models.CASCADE)
    label = models.CharField(max_length=1024)
    lang = models.CharField(
        max_length=4,
        default="en",
        choices=(("en", "English"), ("tet", "Tetum"), ("pt", "Portugese")),
    )



from html.parser import HTMLParser

from formkit_ninja.formkit_schema import DiscriminatedNodeType


class FormKitTagParser(HTMLParser):
    """
    Reverse an HTML example to schema
    This is for lazy copy-pasting from the formkit website :)
    """

    def __init__(self, html_content: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data: str | None = None

        self.current_tag: DiscriminatedNodeType | None = None
        self.tags: list[DiscriminatedNodeType] = []
        self.parents: list[DiscriminatedNodeType] = []
        self.feed(html_content)

    def handle_starttag(self, tag, attrs):
        """
        Read anything that's a "formtag" type
        """
        if tag != "formkit":
            return
        props = dict(attrs)
        props["$formkit"] = props.pop("type")

        tag = DiscriminatedNodeType(**props)
        if isinstance(tag, DiscriminatedNodeType):
            tag = tag.root
        self.current_tag = tag

        if self.parents:
            self.parents[-1].children.append(tag)
        else:
            self.tags.append(tag)
            self.parents.append(tag)

    def handle_endtag(self, tag: str) -> None:
        if tag != "formkit":
            return
        if self.parents:
            self.parents.pop()

    def handle_data(self, data):
        if self.current_tag and data.strip():
            if self.current_tag.children is None:
                self.current_tag.children = [data.strip()]
            else:
                self.current_tag.children.append(data.strip())
            # Ensure that children is included even when "exclude_unset" is True
            # since we populated this after the initial tag build
            self.current_tag.model_fields_set.add("children")



# Generated by Django 4.2.7 on 2023-11-11 04:39

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0024_remove_formkitschemanode_insert_insert_and_more"),
    ]

    operations = [
        pgtrigger.migrations.RemoveTrigger(
            model_name="formkitschemanode",
            name="protect_deletes",
        ),
        pgtrigger.migrations.AddTrigger(
            model_name="formkitschemanode",
            trigger=pgtrigger.compiler.Trigger(
                name="protect_node_deletes_and_updates",
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    condition='WHEN (OLD."protected")',
                    func="RAISE EXCEPTION 'pgtrigger: Cannot delete or update rows from % table', TG_TABLE_NAME;",
                    hash="b1f794b28376aa9aadc4870bcd61260f61c8d8ec",
                    operation="DELETE OR UPDATE",
                    pgid="pgtrigger_protect_node_deletes_and_updates_a71d1",
                    table="formkit_ninja_formkitschemanode",
                    when="BEFORE",
                ),
            ),
        ),
    ]



from django.core.management.base import BaseCommand

from formkit_ninja import models
from formkit_ninja.formkit_schema import FormKitNode, GroupNode
from formkit_ninja.schemas import Schemas


class Command(BaseCommand):
    help = "Load all the Partisipa forms to the database"

    def handle(self, *args, **options):
        models.FormComponents.objects.all().delete()
        models.FormKitSchema.objects.all().delete()
        models.FormKitSchemaNode.objects.all().delete()
        models.Option.objects.all().delete()
        models.OptionGroup.objects.all().delete()

        schemas = Schemas()
        for schema_name in schemas.list_schemas():
            # Each part of the form becomes a 'Schema'
            schema = schemas.as_json(schema_name)
            node: FormKitNode = FormKitNode.parse_obj(schema)
            parsed_node: GroupNode = node.root
            node_in_the_db = list(  # noqa: F841
                models.FormKitSchemaNode.from_pydantic(parsed_node)
            )[
                0
            ]  # noqa: F841



# Generated by Django 4.2.2 on 2023-08-29 08:29

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0010_alter_option_options_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="NodeChildren",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(
                        db_index=True, editable=False, verbose_name="order"
                    ),
                ),
                (
                    "child",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="parent",
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="formkit_ninja.formkitschemanode",
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
            },
        ),
        migrations.AddField(
            model_name="formkitschemanode",
            name="children",
            field=models.ManyToManyField(
                blank=True,
                through="formkit_ninja.NodeChildren",
                to="formkit_ninja.formkitschemanode",
            ),
        ),
    ]



# Generated by Django 4.2.1 on 2023-08-28 05:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "formkit_ninja",
            "0005_option_last_updated_option_object_id_optionlabel_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="option",
            name="group",
            field=models.ForeignKey(
                null=True,
                blank=True,
                default=None,
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.optiongroup",
            ),
            preserve_default=False,
        ),
        migrations.AddConstraint(
            model_name="option",
            constraint=models.UniqueConstraint(
                fields=("group", "object_id"), name="unique_option_id"
            ),
        ),
    ]



# Generated by Django 4.1.7 on 2023-05-08 01:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="formcomponents",
            name="label",
            field=models.CharField(
                help_text="Used as a human-readable label", max_length=1024
            ),
        ),
        migrations.AlterField(
            model_name="formkitschema",
            name="key",
            field=models.TextField(max_length=1024, unique=True),
        ),
    ]



import json
from importlib.resources import files

from formkit_ninja import formkit_schema, samples


def test_node_parse():
    schema = json.loads(files(samples).joinpath("element.json").read_text())
    formkit_schema.DiscriminatedNodeType.model_validate(schema[0])


def test_login_form():
    schema = json.loads(
        files(samples).joinpath("form_generation_example.json").read_text()
    )
    formkit_schema.DiscriminatedNodeType.model_validate(schema[0])


def test_raw_values():
    schema = json.loads(files(samples).joinpath("raw_values.json").read_text())
    node = formkit_schema.DiscriminatedNodeType.model_validate(schema)
    node_reloaded = json.loads(node.model_dump_json(by_alias=True, exclude_unset=True))
    assert node_reloaded["props"]["__raw__price"] == "$2.99"
    assert node_reloaded["$cmp"] == "PriceComponent"


def test_meeting_type_node():
    """
    Loads a single element schema, checking that return values
    are the same as entered values
    """
    schema = json.loads(files(samples).joinpath("meeting_type_node.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])
    meeting_type_schema = formkit_schema.FormKitSchema.parse_obj(schema)
    reloaded = json.loads(
        meeting_type_schema.model_dump_json(by_alias=True, exclude_none=True)
    )

    assert reloaded[0].get("id") == "meeting_type"
    assert reloaded[0].get("name") == "meeting_type"
    assert reloaded[0].get("$formkit") == "select"

    assert len(reloaded[0]["options"]) == 2


def test_repeater():
    schema = json.loads(files(samples).joinpath("repeater.json").read_text())
    model = formkit_schema.FormKitNode.model_validate(schema[0])
    assert model.root.model_dump(by_alias=True, exclude_none=True)


def test_dropdown():
    schema = json.loads(files(samples).joinpath("dropdown.json").read_text())
    formkit_schema.FormKitNode.parse_obj(schema[0])


def test_formkit_element():
    schema = {
        "$formkit": "group",
        "name": "Top level",
        "children": [
            {
                "$formkit": "group",
                "name": "2nd level",
                "children": [
                    {
                        "$formkit": "group",
                        "name": "3rd level",
                        "children": [
                            {
                                "$formkit": "group",
                                "name": "4th level level",
                            }
                        ],
                    }
                ],
            }
        ],
    }
    validated = formkit_schema.FormKitNode.model_validate(schema)
    object = validated.root.model_dump(by_alias=True, exclude_none=True)
    assert validated.root.node_type == "formkit"
    assert isinstance(validated.root, formkit_schema.GroupNode)
    assert isinstance(validated.root.children[0], formkit_schema.GroupNode)
    assert object.keys() == schema.keys()
    assert schema == object


def test_by_alias():
    """
    Some fields have an alias.
    Generally where html attributes are hyphenated, such as `validation-label`
    which is aliased to `validationLabel`
    """
    PasswordNode = formkit_schema.PasswordNode
    passnode = PasswordNode(**{"validation-label": "Password"})
    passnode_upper = PasswordNode(**{"validationLabel": "Password"})
    assert passnode.validationLabel == "Password"
    assert passnode.validationLabel == passnode_upper.validationLabel



import json
from importlib.resources import files

import pytest

from formkit_ninja import formkit_schema, models, samples
from formkit_ninja.formkit_schema import (DiscriminatedNodeType, FormKitNode,
                                          FormKitSchemaDOMNode, SelectNode)
from formkit_ninja.schemas import Schemas

schemas = Schemas().schemas


def _read_file(filename: str):
    return json.loads(files(samples).joinpath(f"{filename}.json").read_text())


def _parse_file(filename: str):
    return formkit_schema.FormKitSchema.model_validate(_read_file(filename))


@pytest.fixture(params=list(schemas))
def schema(request):
    return s.as_json(request.param)


def test_node_parse():
    schema = json.loads(files(samples).joinpath("element.json").read_text())
    formkit_schema.FormKitNode.model_validate(schema[0])
    _parse_file("element")


@pytest.mark.django_db()
def test_create_schema():
    c = models.FormKitSchema.objects.create()  # noqa: F841


@pytest.mark.django_db()
def test_create_from_schema():
    """
    Create a database entry from a FormKit schema
    """
    element_schema = formkit_schema.FormKitSchema.model_validate(
        [
            {
                "$el": "div",
                "attrs": {"style": {"color": "red"}, "data-foo": "bar"},
                "children": "Hello world",
            }
        ]
    )
    assert element_schema.root[0].node_type == "element"
    c = models.FormKitSchema.from_pydantic(element_schema)
    # This test schema has one node
    assert c.nodes.count() == 1
    # The "Hello world" is one node
    hello_world_text = c.nodes.first().children.first()
    assert hello_world_text.text_content == "Hello world"

    # The node...
    first_node = c.nodes.first()
    dictify = first_node.get_node(recursive=True).model_dump(
        exclude_none=True, by_alias=True
    )

    # The node looks like this:
    # {'children': [], 'node_type': 'element', 'el': 'div', 'attrs': {'style': {...}, 'data-foo': 'bar'}}

    assert set(dictify.keys()) == {"children", "$el", "attrs"}
    assert dictify["children"] == ["Hello world"]
    assert dictify["$el"] == "div"
    assert set(dictify["attrs"].keys()) == {"style", "data-foo"}

    # A 'group' node
    first_node = c.nodes.first()
    first_node.get_node().model_dump(
        exclude_none=True, by_alias=True, exclude={"children"}
    )


@pytest.mark.django_db()
def test_parse_el_priority(el_priority: dict):
    """
    Regression test for missing text content

    Expects: {
        "$el": "div",
        "children": ["Priority ", {"$el": "span", "attrs": {"class": "ml-1"}, "children": ["$: ($index + 1)"]}],
        "attrs": {"class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"},
    }
    """
    parsed_node = DiscriminatedNodeType.model_validate(el_priority).root
    assert isinstance(parsed_node, FormKitSchemaDOMNode)
    # Load it into the database
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]
    # The node type in the database is `$el`
    assert node_in_the_db.node_type == "$el"

    el_priority["children"][0] == "Priority "
    children: list[models.FormKitSchemaNode] = list(
        node_in_the_db.children.all().order_by("nodechildren__order")
    )

    # The first child is an 'text'
    node = children[0]
    assert node.node_type == "text"
    assert node.text_content == "Priority "
    assert node.to_pydantic() == "Priority "

    # The second child is an 'el'
    node = children[1]
    assert node.node_type == "$el"
    assert node.node.get("$el") == "span"
    assert node.node["attrs"] == {"class": "ml-1"}
    # With appropriate `by_alias` and `exclude_defaults` we should get an equal output as input
    assert (
        node.to_pydantic(recursive=True).model_dump(
            by_alias=True, exclude_defaults=True
        )
        == el_priority["children"][1]
    )


@pytest.mark.django_db()
def test_children_string():
    """
    Test that a node with children as a string is parsed correctly
    """
    node = FormKitNode.model_validate(
        {
            "$el": "div",
            "children": "Hello world",
            "attrs": {
                "class": "rounded-full px-5 py-2 bg-zinc-400 text-lg font-bold mb-5"
            },
        }
    ).root
    assert node.children == "Hello world"


@pytest.mark.django_db()
def test_parse_simple_text_node(simple_text_node: dict):
    node: FormKitNode = FormKitNode.model_validate(simple_text_node)
    parsed_node: FormKitSchemaDOMNode = node.root
    # Before the fix, this was breaking into individual letters
    assert parsed_node.children == simple_text_node["children"] == "Priority"


@pytest.mark.django_db()
def test_nested_additional(nested_formkit_text_node: dict):
    node: FormKitNode = DiscriminatedNodeType.model_validate(nested_formkit_text_node)

    # From JSON to a Pydantic class...
    # Because "class" is not a standard attribute, it is stored in the
    # 'additional_props' field
    parsed_node: formkit_schema.SelectNode = node.root  # noqa: F841
    assert nested_formkit_text_node["children"][0]["class"] == "red"
    assert node.root.children[0].additional_props == {"class": "red"}
    assert "additional_props" not in node.root.model_dump()["children"][0]
    assert "class" in node.root.model_dump()["children"][0]
    assert (
        node.root.model_dump(by_alias=True, exclude_none=True)
        == nested_formkit_text_node
    )


@pytest.mark.django_db()
def test_additional_props(formkit_text_node: dict):
    """
    This ensures that custom attributes are maintained
    from JSON to Pydantic,
    from Pydantic to the Database,
    from the Database to Pydantic,
    and from Pydantic to JSON.
    """
    assert formkit_text_node["class"] == "red"
    node: FormKitNode = FormKitNode.model_validate(formkit_text_node)

    # From JSON to a Pydantic class...
    # Because "class" is not a standard attribute, it is stored in the
    # 'additional_props' field
    parsed_node: formkit_schema.SelectNode = node.root
    assert parsed_node.additional_props == {"class": "red"}

    # Into the database...
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]
    assert node_in_the_db.additional_props == {"class": "red"}
    node_in_the_db.to_pydantic()

    # The 'discriminator' fields should not be stored in the db
    assert node_in_the_db.node["$formkit"] == "select"
    assert set(node_in_the_db.node.keys()) == {
        "key",
        "id",
        "name",
        "label",
        "placeholder",
        "$formkit",
    }
    # And out of the database again

    select_node = FormKitNode.model_validate(node_in_the_db.node).root
    assert node_in_the_db.additional_props == {"class": "red"}

    assert isinstance(select_node, SelectNode)

    # When we've pulled this out from the database, we did not include "additional_props"
    assert select_node.key == "activity_type"
    assert select_node.id == "activity_type"
    assert select_node.name == "activity_type"

    # Now we'll do the same but we include "additional_props"
    select_node = FormKitNode.model_validate(
        {
            **node_in_the_db.node,
            "additional_props": node_in_the_db.additional_props,
            "options": node_in_the_db.node_options,
        }
    ).root

    assert select_node.additional_props == {"class": "red"}
    assert select_node.model_dump(by_alias=True, exclude_none=True)["class"] == "red"

    from_the_db = node_in_the_db.to_pydantic(options=True)

    assert from_the_db.root.additional_props == {"class": "red"}
    assert from_the_db.root.model_dump()["class"] == "red"

    # And back to JSON
    json_from_the_db = from_the_db.model_dump(
        by_alias=True, exclude_none=True, exclude={"node_type"}
    )
    assert json_from_the_db["class"] == "red"

    # Additional checks that the JSON output is equivalent to the JSON input
    # Note that json from the db has additional Python 'discriminator' fields 'node_type' and 'formkit'

    assert set(node_in_the_db.node.keys()) == {
        "key",
        "id",
        "name",
        "label",
        "placeholder",
        "$formkit",
    }

    assert node_in_the_db.node.get("key") == formkit_text_node["key"]
    assert node_in_the_db.node.get("id") == formkit_text_node["id"]
    assert node_in_the_db.node.get("name") == formkit_text_node["name"]
    assert node_in_the_db.node.get("label") == formkit_text_node["label"]
    assert node_in_the_db.node.get("$formkit") == formkit_text_node["$formkit"]
    assert node_in_the_db.node.get("placeholder") == formkit_text_node["placeholder"]
    assert node_in_the_db.additional_props == {"class": "red"}

    assert json_from_the_db["key"] == formkit_text_node["key"]
    assert json_from_the_db["id"] == formkit_text_node["id"]
    assert json_from_the_db["name"] == formkit_text_node["name"]
    assert json_from_the_db["label"] == formkit_text_node["label"]
    assert json_from_the_db["$formkit"] == formkit_text_node["$formkit"]
    assert json_from_the_db["placeholder"] == formkit_text_node["placeholder"]
    # This "additional prop" is now merged again to the json data
    assert json_from_the_db["class"] == formkit_text_node["class"]
    assert json_from_the_db["options"] == formkit_text_node["options"]


s = Schemas()
schemas = s.list_schemas()


def schema_are_same(in_: dict | str, out_: dict | str):
    if isinstance(in_, str):
        assert in_ == out_, f"{in_=} != {out_=}"
        return
    elif isinstance(in_, dict) and isinstance(out_, dict):
        assert set(in_.keys()) == set(out_.keys())
        for key in in_.keys():
            if key in {"options", "children"}:
                continue
            assert in_[key] == out_[key], f"{in_[key]=} != {out_[key]=}"
    if "children" in in_ and isinstance(out_, dict):
        for c_in, c_out in zip(in_.get("children", []), out_.get("children", [])):
            schema_are_same(c_in, c_out)


@pytest.mark.django_db()
def test_schemas(schema: dict):
    m = DiscriminatedNodeType.model_validate(schema)
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(m.root))[0]
    schema_from_db: FormKitNode = node_in_the_db.to_pydantic(
        recursive=True, options=True
    )

    schema_out = schema_from_db.model_dump(by_alias=True, exclude_none=True)

    schema_are_same(schema_out, schema)

    return

    import pyinstrument

    profiler = pyinstrument.Profiler()
    profiler.start()

    node: FormKitNode = FormKitNode.model_validate(schema)
    profiler.stop()
    profiler.print()

    parsed_node: formkit_schema.SelectNode = node.root
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]

    # Returning the code
    # Database model to pydantic
    schema_from_db: FormKitNode = node_in_the_db.to_pydantic(
        recursive=True, options=True
    )

    # assert schema_from_db.root.children[0].model_dump()["$formkit"] == "group"
    # assert schema_from_db.root.children[1].children[1].if_condition == "$get(sector_id_dropdown).value"

    schema_out: dict = schema_from_db.model_dump(by_alias=True, exclude_none=True)

    assert schema.keys() == schema_out.keys()
    schema_are_same(schema, schema_out)


def test_if_condition():
    schema = {
        "$formkit": "radio",
        "id": "subsector_id",
        "if": "$get(sector_id).value",
        "key": "subsector",
        "label": "$gettext(Subsector)",
        "name": "subsector",
        "options": '$ida(subsector, "sector_id="+$get(sector_id).value)',
    }

    node: FormKitNode = FormKitNode.model_validate(schema)

    assert node.model_dump(by_alias=True, exclude_none=True) == schema

    assert node.root.if_condition == "$get(sector_id).value"
    assert "if_condition" not in node.model_dump()
    assert node.model_dump()["if"] == "$get(sector_id).value"


def test_if_group_condition():
    schema = {
        "$formkit": "group",
        "children": [
            {
                "$formkit": "radio",
                "id": "subsector_id",
                "if": "$get(sector_id).value",
                "key": "subsector",
                "label": "$gettext(Subsector)",
                "name": "subsector",
                "options": '$ida(subsector, "sector_id="+$get(sector_id).value)',
            }
        ],
    }

    node: FormKitNode = FormKitNode.model_validate(schema)
    # assert node.root.children[0].if_condition == "$get(sector_id).value"
    node.root.model_dump()
    schema_out = node.model_dump(by_alias=True, exclude_none=True)
    assert schema_out == schema
    # assert node.model_dump()['children'][0]["if"] == "$get(sector_id).value"


@pytest.mark.django_db()
def test_protected_model(formkit_text_node: dict):
    """
    A 'protected' node cannot be deleted
    This works right at the database level, not through the API
    We can test directly with a call to `.delete`
    """
    node: FormKitNode = FormKitNode.model_validate(formkit_text_node)
    parsed_node: formkit_schema.SelectNode = node.root
    node_in_the_db = list(models.FormKitSchemaNode.from_pydantic(parsed_node))[0]
    node_in_the_db.protected = True
    node_in_the_db.save()
    from django.db.utils import InternalError

    with pytest.raises(InternalError):
        node_in_the_db.delete()



# Generated by Django 4.2.4 on 2023-09-04 10:35

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "formkit_ninja",
            "0012_remove_formkitschemanode_group_remove_option_field_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="formkitschemanode",
            name="option_group",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="formkit_ninja.optiongroup",
            ),
        ),
    ]



# Generated by Django 4.2.1 on 2023-08-28 06:13

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("formkit_ninja", "0008_remove_option_label"),
    ]

    operations = [
        migrations.AddField(
            model_name="formkitschemanode",
            name="text_content",
            field=models.TextField(
                blank=True,
                help_text="Content for a text element, for children of an $el type component",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="nodechildren",
            name="parent",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="formkit_ninja.formkitschemanode",
            ),
        ),
    ]



from formkit_ninja.api import disambiguate_name, make_name_valid_id


def test_disambiguate_name():
    assert disambiguate_name("test", set()) == "test"
    assert disambiguate_name("test", {"test"}) == "test_1"
    assert disambiguate_name("test", {"test_1"}) == "test"
    assert disambiguate_name("test", {"test", "test_1", "test_2"}) == "test_3"


def test_make_name_valid_id():
    assert make_name_valid_id("1Foo_") == "_1foo"
    assert make_name_valid_id("Foo_") == "foo"
    assert make_name_valid_id("Foo1") == "foo1"

    # Resolves an issue where an invalid char followed by underscore creates a 'bad' name
    assert make_name_valid_id("01Foo?_") == "_01foo"



from __future__ import annotations

import warnings
from keyword import iskeyword
from typing import Iterable, Literal

from formkit_ninja import formkit_schema
from formkit_ninja.formkit_schema import FormKitNode, GroupNode, RepeaterNode

FormKitType = formkit_schema.FormKitType


def make_valid_identifier(input_string: str):
    """
    Replace invalid characters with underscores
    Remove trailing / leading digits
    Remove trailing/leading underscores
    Lowercase
    """
    try:
        output = "".join(ch if ch.isalnum() else "_" for ch in input_string)

        while output[-1].isdigit():
            output = output[:-1]

        while output[0].isdigit():
            output = output[1:]

        while output[-1] == "_":
            output = output[:-1]

        while output[0] == "_":
            output = output[1:]
    except IndexError:
        raise TypeError(f"The name {input_string} couldn't be used as an identifier")

    return output.lower()


class NodePath:
    """
    Mostly a wrapper around "tuple" to provide useful conventions
    for naming
    """

    def __init__(self, *nodes: FormKitType):
        self.nodes = nodes

    @classmethod
    def from_obj(cls, obj: dict):
        return cls(FormKitNode.model_validate(obj).root)

    def __truediv__(self, node: Literal[".."] | FormKitType):
        """
        This overrides the builtin '/' operator, like "Path", to allow appending nodes
        """
        if node == "..":
            return self.__class__(*self.nodes[:-1])
        return self.__class__(*self.nodes, node)

    def suggest_model_name(self) -> str:
        """
        Single reference for table name and foreign key references
        """
        model_name = "".join(map(self.safe_node_name, self.nodes))
        return model_name

    def suggest_class_name(self):
        model_name = "".join(
            map(lambda n: n.capitalize(), map(self.safe_node_name, self.nodes))
        )
        return model_name

    def suggest_field_name(self):
        """
        Single reference for table name and foreign key references
        """
        return self.safe_node_name(self.node)

    def suggest_link_class_name(self):
        return f"{self.classname}Link"

    # Some accessors for the functions above

    @property
    def modelname(self):
        return self.suggest_model_name()

    @property
    def classname(self):
        return self.suggest_class_name()

    @property
    def fieldname(self):
        return self.suggest_field_name()

    @property
    def linkname(self):
        return self.suggest_link_class_name()

    @property
    def classname_lower(self):
        return self.classname.lower()

    @property
    def classname_schema(self):
        return f"{self.classname}Schema"

    @staticmethod
    def safe_name(name: str, fix: bool = True) -> str:
        """
        Ensure that the "name" provided is a valid
        python identifier, correct if necessary
        """
        if name is None:
            raise TypeError
        if not name.isidentifier() or iskeyword(name):
            if fix:
                warnings.warn(f"The name:  '''{name}''' is not a valid identifier")
                # Run again to check that it's not a keyword
                return NodePath.safe_name(make_valid_identifier(name), fix=False)
            else:
                raise KeyError(f"The name:  '''{name}''' is not a valid identifier")
        return name

    def safe_node_name(self, node: FormKitType) -> str:
        """
        Return either the "name" or "id" field
        """
        if node.name:
            name = self.safe_name(node.name)
        elif node.id:
            name = self.safe_name(node.id)
        else:
            raise AttributeError("Could not determine a suitable 'name' for this node")

        return name

    @property
    def is_repeater(self):
        return isinstance(self.node, RepeaterNode)

    @property
    def is_group(self):
        return isinstance(self.node, GroupNode)

    @property
    def formkits(self) -> Iterable["NodePath"]:
        for n in self.children:
            if hasattr(n, "formkit"):
                yield self / n

    @property
    def formkits_not_repeaters(self) -> tuple["NodePath"]:
        def _get():
            for n in self.children:
                if hasattr(n, "formkit") and not isinstance(n, RepeaterNode):
                    yield self / n

        return tuple(_get())

    @property
    def children(self):
        return getattr(self.node, "children", []) or []

    def filter_children(self, type_) -> Iterable["NodePath"]:
        for n in self.children:
            if isinstance(n, type_):
                yield self / n

    @property
    def repeaters(self):
        return tuple(self.filter_children(RepeaterNode))

    @property
    def groups(self):
        return tuple(self.filter_children(GroupNode))

    @property
    def node(self):
        return self.nodes[-1]

    @property
    def parent(self):
        if len(self.nodes) > 1:
            return self.nodes[-2]
        else:
            return None

    @property
    def is_child(self):
        return self.parent is not None

    @property
    def depth(self):
        return len(self.nodes)

    @property
    def tail(self):
        return NodePath(self.node)

    def __str__(self):
        return f"NodePath {len(self.nodes)}: {self.node}"

    @property
    def django_attrib_name(self):
        """
        If not a group, return the Django field attribute
        """
        return self.tail.modelname

    @property
    def pydantic_attrib_name(self):
        base = self.django_attrib_name
        return base

    @property
    def parent_class_name(self):
        return (self / "..").classname

    def to_pydantic_type(
        self,
    ) -> Literal["str", "int", "bool", "Decimal", "float", "date"] | str:
        """
        Usually, this should return a well known Python type as a string
        """
        node = self.node
        if node.formkit == "number":
            if node.step is not None:
                # We don't actually **know** this but it's a good assumption
                return "float"
            return "int"

        match node.formkit:
            case "text":
                return "str"
            case "number":
                return "float"
            case "select" | "dropdown" | "radio" | "autocomplete":
                return "str"
            case "datepicker":
                return "datetime"
            case "tel":
                return "int"
            case "group":
                return self.classname
            case "repeater":
                return f"list[{self.classname}]"
            case "hidden":
                return "str"
        return "str"

    @property
    def pydantic_type(self):
        return self.to_pydantic_type()

    def to_postgres_type(self):
        match self.to_pydantic_type():
            case "bool":
                return "boolean"
            case "str":
                return "text"
            case "Decimal":
                return "NUMERIC(15,2)"
            case "int":
                return "int"
            case "float":
                return "float"
        return "text"

    @property
    def postgres_type(self):
        return self.to_postgres_type()

    def to_django_type(self) -> str:
        """
        Return the "models.ModelField" which would match this data type
        """
        if self.is_group:
            return "OneToOneField"

        match self.to_pydantic_type():
            case "bool":
                return "BooleanField"
            case "str":
                return "TextField"
            case "Decimal":
                return "DecimalField"
            case "int":
                return "IntegerField"
            case "float":
                return "FloatField"
            case "datetime":
                return "DateTimeField"
            case "date":
                return "DateField"
            case "UUID":
                return "UUIDField"
        return "TextField"

    @property
    def django_type(self):
        return self.to_django_type()

    def to_django_args(self) -> str:
        if self.is_group:
            return f"{self.classname}, on_delete=models.CASCADE"

        match self.to_pydantic_type():
            case "bool":
                return "null=True, blank=True"
            case "str":
                return "null=True, blank=True"
            case "Decimal":
                return "max_digits=20, decimal_places=2, null=True, blank=True"
            case "int":
                return "null=True, blank=True"
            case "float":
                return "null=True, blank=True"
            case "datetime":
                return "null=True, blank=True"
            case "date":
                return "null=True, blank=True"
            case "UUID":
                return "editable=False, null=True, blank=True"
        raise TypeError("Undetermined Django field type")

    @property
    def django_args(self):
        return self.to_django_args()

    @property
    def extra_attribs(self):
        """
        Returns extra fields to be appended to this group or
        repeater node in "models.py"
        """
        return []

    @property
    def extra_attribs_schema(self):
        """
        Returns extra attributes to be appended to "schema_out.py"
        For Partisipa this included a foreign key to "Submission"
        """
        return []

    @property
    def extra_attribs_basemodel(self):
        """
        Returns extra attributes to be appended to "schema.py"
        For Partisipa this included a foreign key to "Submission"
        """
        return []

    @property
    def validators(self) -> list[str]:
        """
        Hook to allow extra processing for
        fields like Partisipa's 'currency' field
        """
        return []
